# Copied and modified from Matplotlib
import ast
import contextlib
import doctest
import itertools
import os
import re
import shutil
import sys
import textwrap
import traceback
from io import StringIO
from os.path import relpath
from pathlib import Path

import jinja2
import matplotlib
import matplotlib.pyplot as plt
from cycler import cycler
from docutils.parsers.rst import Directive, directives
from docutils.parsers.rst.directives.images import Image
from matplotlib import _pylab_helpers, cbook
from matplotlib.backend_bases import FigureManagerBase
from matplotlib.colors import ListedColormap
from sphinx.errors import ExtensionError

_WBCOLORMAP = ListedColormap(
    [
        [0.700151, 0.002745, 0.700612],
        [0.700191, 0.010833, 0.697186],
        [0.700226, 0.019196, 0.693784],
        [0.700255, 0.027497, 0.690410],
        [0.700279, 0.036129, 0.687067],
        [0.700299, 0.044535, 0.683750],
        [0.700315, 0.052201, 0.680467],
        [0.700328, 0.059479, 0.677234],
        [0.700339, 0.066138, 0.674024],
        [0.700347, 0.072500, 0.670865],
        [0.700355, 0.078557, 0.667747],
        [0.700362, 0.084489, 0.664665],
        [0.700369, 0.090118, 0.661647],
        [0.700378, 0.095602, 0.658662],
        [0.700389, 0.100919, 0.655740],
        [0.700403, 0.106180, 0.652854],
        [0.700422, 0.111272, 0.650035],
        [0.700447, 0.116281, 0.647260],
        [0.700479, 0.121141, 0.644532],
        [0.700521, 0.126029, 0.641864],
        [0.700575, 0.130794, 0.639246],
        [0.700642, 0.135455, 0.636684],
        [0.700726, 0.140079, 0.634169],
        [0.700829, 0.144685, 0.631701],
        [0.700953, 0.149193, 0.629284],
        [0.701102, 0.153641, 0.626923],
        [0.701279, 0.158084, 0.624615],
        [0.701486, 0.162466, 0.622349],
        [0.701729, 0.166801, 0.620147],
        [0.702012, 0.171076, 0.617979],
        [0.702337, 0.175336, 0.615867],
        [0.702707, 0.179555, 0.613800],
        [0.703122, 0.183741, 0.611790],
        [0.703582, 0.187936, 0.609823],
        [0.704093, 0.192033, 0.607911],
        [0.704658, 0.196154, 0.606029],
        [0.705275, 0.200200, 0.604213],
        [0.705946, 0.204298, 0.602424],
        [0.706671, 0.208310, 0.600685],
        [0.707439, 0.212332, 0.598992],
        [0.708265, 0.216333, 0.597337],
        [0.709140, 0.220302, 0.595714],
        [0.710053, 0.224219, 0.594134],
        [0.711015, 0.228186, 0.592595],
        [0.712011, 0.232099, 0.591073],
        [0.713044, 0.235997, 0.589585],
        [0.714113, 0.239864, 0.588131],
        [0.715214, 0.243717, 0.586706],
        [0.716339, 0.247586, 0.585309],
        [0.717485, 0.251421, 0.583920],
        [0.718642, 0.255233, 0.582557],
        [0.719825, 0.259028, 0.581216],
        [0.721012, 0.262798, 0.579897],
        [0.722212, 0.266557, 0.578580],
        [0.723423, 0.270315, 0.577284],
        [0.724641, 0.274034, 0.576009],
        [0.725860, 0.277747, 0.574723],
        [0.727075, 0.281433, 0.573468],
        [0.728294, 0.285107, 0.572216],
        [0.729516, 0.288786, 0.570967],
        [0.730731, 0.292428, 0.569729],
        [0.731948, 0.296063, 0.568502],
        [0.733159, 0.299689, 0.567287],
        [0.734368, 0.303280, 0.566067],
        [0.735568, 0.306883, 0.564853],
        [0.736769, 0.310471, 0.563661],
        [0.737963, 0.314024, 0.562464],
        [0.739164, 0.317574, 0.561261],
        [0.740347, 0.321115, 0.560077],
        [0.741527, 0.324632, 0.558885],
        [0.742714, 0.328153, 0.557711],
        [0.743886, 0.331654, 0.556532],
        [0.745056, 0.335166, 0.555357],
        [0.746225, 0.338633, 0.554196],
        [0.747393, 0.342111, 0.553026],
        [0.748552, 0.345565, 0.551874],
        [0.749704, 0.349024, 0.550714],
        [0.750868, 0.352465, 0.549561],
        [0.752015, 0.355908, 0.548416],
        [0.753158, 0.359324, 0.547260],
        [0.754311, 0.362743, 0.546124],
        [0.755451, 0.366162, 0.544977],
        [0.756589, 0.369556, 0.543837],
        [0.757726, 0.372949, 0.542705],
        [0.758860, 0.376350, 0.541561],
        [0.759990, 0.379722, 0.540437],
        [0.761122, 0.383103, 0.539305],
        [0.762246, 0.386459, 0.538189],
        [0.763371, 0.389823, 0.537065],
        [0.764497, 0.393183, 0.535933],
        [0.765621, 0.396533, 0.534816],
        [0.766735, 0.399878, 0.533691],
        [0.767850, 0.403228, 0.532586],
        [0.768965, 0.406563, 0.531468],
        [0.770080, 0.409898, 0.530357],
        [0.771197, 0.413225, 0.529252],
        [0.772302, 0.416545, 0.528136],
        [0.773408, 0.419864, 0.527036],
        [0.774514, 0.423171, 0.525928],
        [0.775620, 0.426487, 0.524822],
        [0.776727, 0.429800, 0.523723],
        [0.777826, 0.433105, 0.522621],
        [0.778921, 0.436413, 0.521521],
        [0.780016, 0.439713, 0.520427],
        [0.781110, 0.443016, 0.519342],
        [0.782202, 0.446308, 0.518256],
        [0.783285, 0.449611, 0.517171],
        [0.784373, 0.452911, 0.516077],
        [0.785450, 0.456202, 0.515007],
        [0.786526, 0.459505, 0.513939],
        [0.787590, 0.462792, 0.512865],
        [0.788657, 0.466096, 0.511800],
        [0.789713, 0.469411, 0.510744],
        [0.790763, 0.472703, 0.509696],
        [0.791807, 0.476015, 0.508661],
        [0.792845, 0.479325, 0.507622],
        [0.793877, 0.482635, 0.506589],
        [0.794892, 0.485949, 0.505565],
        [0.795900, 0.489268, 0.504572],
        [0.796901, 0.492594, 0.503560],
        [0.797894, 0.495924, 0.502571],
        [0.798879, 0.499261, 0.501599],
        [0.799853, 0.502601, 0.500622],
        [0.800811, 0.505942, 0.499654],
        [0.801763, 0.509301, 0.498715],
        [0.802706, 0.512652, 0.497778],
        [0.803641, 0.516005, 0.496832],
        [0.804568, 0.519378, 0.495907],
        [0.805481, 0.522745, 0.495000],
        [0.806395, 0.526120, 0.494092],
        [0.807296, 0.529496, 0.493183],
        [0.808190, 0.532873, 0.492294],
        [0.809081, 0.536254, 0.491419],
        [0.809960, 0.539633, 0.490532],
        [0.810839, 0.543029, 0.489643],
        [0.811716, 0.546423, 0.488777],
        [0.812583, 0.549811, 0.487916],
        [0.813449, 0.553199, 0.487046],
        [0.814314, 0.556601, 0.486178],
        [0.815178, 0.560003, 0.485315],
        [0.816042, 0.563405, 0.484456],
        [0.816898, 0.566806, 0.483600],
        [0.817752, 0.570200, 0.482748],
        [0.818608, 0.573611, 0.481898],
        [0.819464, 0.577019, 0.481035],
        [0.820320, 0.580426, 0.480189],
        [0.821177, 0.583832, 0.479337],
        [0.822036, 0.587241, 0.478479],
        [0.822888, 0.590654, 0.477626],
        [0.823741, 0.594071, 0.476774],
        [0.824596, 0.597489, 0.475924],
        [0.825452, 0.600903, 0.475075],
        [0.826310, 0.604330, 0.474213],
        [0.827167, 0.607752, 0.473350],
        [0.828020, 0.611173, 0.472504],
        [0.828883, 0.614590, 0.471655],
        [0.829740, 0.618022, 0.470796],
        [0.830598, 0.621451, 0.469943],
        [0.831459, 0.624886, 0.469087],
        [0.832315, 0.628325, 0.468218],
        [0.833174, 0.631759, 0.467368],
        [0.834038, 0.635197, 0.466506],
        [0.834897, 0.638641, 0.465638],
        [0.835759, 0.642088, 0.464790],
        [0.836622, 0.645532, 0.463931],
        [0.837488, 0.648990, 0.463055],
        [0.838348, 0.652439, 0.462201],
        [0.839214, 0.655902, 0.461342],
        [0.840083, 0.659357, 0.460469],
        [0.840947, 0.662826, 0.459614],
        [0.841815, 0.666286, 0.458749],
        [0.842684, 0.669764, 0.457878],
        [0.843548, 0.673236, 0.457012],
        [0.844421, 0.676714, 0.456145],
        [0.845291, 0.680187, 0.455282],
        [0.846160, 0.683675, 0.454415],
        [0.847033, 0.687166, 0.453536],
        [0.847907, 0.690654, 0.452675],
        [0.848784, 0.694149, 0.451800],
        [0.849662, 0.697652, 0.450924],
        [0.850534, 0.701158, 0.450049],
        [0.851410, 0.704667, 0.449178],
        [0.852296, 0.708180, 0.448296],
        [0.853170, 0.711699, 0.447413],
        [0.854049, 0.715220, 0.446541],
        [0.854936, 0.718742, 0.445655],
        [0.855818, 0.722269, 0.444788],
        [0.856699, 0.725811, 0.443905],
        [0.857584, 0.729347, 0.443021],
        [0.858470, 0.732893, 0.442142],
        [0.859359, 0.736436, 0.441247],
        [0.860250, 0.739997, 0.440361],
        [0.861136, 0.743552, 0.439474],
        [0.862032, 0.747115, 0.438584],
        [0.862924, 0.750684, 0.437706],
        [0.863818, 0.754257, 0.436810],
        [0.864708, 0.757833, 0.435919],
        [0.865604, 0.761413, 0.435028],
        [0.866502, 0.765006, 0.434120],
        [0.867400, 0.768593, 0.433228],
        [0.868307, 0.772195, 0.432343],
        [0.869206, 0.775798, 0.431443],
        [0.870103, 0.779408, 0.430538],
        [0.871012, 0.783023, 0.429635],
        [0.871923, 0.786651, 0.428733],
        [0.872831, 0.790273, 0.427834],
        [0.873742, 0.793911, 0.426929],
        [0.874658, 0.797545, 0.426015],
        [0.875582, 0.801192, 0.425119],
        [0.876504, 0.804847, 0.424205],
        [0.877438, 0.808510, 0.423294],
        [0.878381, 0.812177, 0.422386],
        [0.879330, 0.815859, 0.421480],
        [0.880295, 0.819541, 0.420578],
        [0.881278, 0.823243, 0.419683],
        [0.882276, 0.826956, 0.418775],
        [0.883302, 0.830680, 0.417880],
        [0.884356, 0.834420, 0.416995],
        [0.885442, 0.838174, 0.416101],
        [0.886573, 0.841950, 0.415222],
        [0.887744, 0.845741, 0.414352],
        [0.888981, 0.849563, 0.413497],
        [0.890275, 0.853401, 0.412658],
        [0.891647, 0.857274, 0.411817],
        [0.893098, 0.861176, 0.411008],
        [0.894649, 0.865109, 0.410221],
        [0.896296, 0.869084, 0.409443],
        [0.898055, 0.873088, 0.408695],
        [0.899941, 0.877132, 0.407976],
        [0.901952, 0.881224, 0.407286],
        [0.904103, 0.885355, 0.406631],
        [0.906404, 0.889529, 0.406013],
        [0.908850, 0.893747, 0.405420],
    ],
    name="wildboar",
)
try:
    matplotlib.colormaps.register(_WBCOLORMAP)
except:
    pass


def light_dark_paths(path: Path):
    file = Path(path.parts[-1])
    path = Path(*path.parts[:-1])
    ext = file.suffix
    name = file.stem

    return path.joinpath(f"{name}-light{ext}"), path.joinpath(f"{name}-dark{ext}")


def interleave(b):
    o = b[::2]
    if len(b) % 2 == 0:
        start = -1
    else:
        start = -2
    r = b[start::-2]
    oj = 0
    rj = 0
    new = []
    for i in range(len(b)):
        if i % 2 == 0:
            new.append(o[oj])
            oj += 1
        else:
            new.append(r[rj])
            rj += 1
    return new


colors = interleave(
    [
        matplotlib.colors.rgb2hex(rgb)
        for rgb in matplotlib.colormaps["wildboar"].resampled(10).colors
    ]
)


matplotlib.use("agg")

__version__ = 2


# -----------------------------------------------------------------------------
# Registration hook
# -----------------------------------------------------------------------------


def _option_boolean(arg):
    if not arg or not arg.strip():
        # no argument given, assume used as a flag
        return True
    elif arg.strip().lower() in ("no", "0", "false"):
        return False
    elif arg.strip().lower() in ("yes", "1", "true"):
        return True
    else:
        raise ValueError(f"{arg!r} unknown boolean")


def _option_context(arg):
    if arg in [None, "reset", "close-figs"]:
        return arg
    raise ValueError("Argument should be None or 'reset' or 'close-figs'")


def _option_format(arg):
    return directives.choice(arg, ("python", "doctest"))


def _option_output_format(arg):
    if arg is None:
        return "plain"
    return directives.choice(arg, ("code", "plain"))


def _option_card_width(arg):
    return directives.choice(arg, ("auto", "25%", "50%", "75%", "100%"))


def _option_carousel(arg):
    return directives.choice(arg, (1, 2, 3))


def mark_plot_labels(app, document):
    """
    To make plots referenceable, we need to move the reference from the
    "htmlonly" (or "latexonly") node to the actual figure node itself.
    """
    for name, explicit in document.nametypes.items():
        if not explicit:
            continue
        labelid = document.nameids[name]
        if labelid is None:
            continue
        node = document.ids[labelid]
        if node.tagname in ("html_only", "latex_only"):
            for n in node:
                if n.tagname == "figure":
                    sectname = name
                    for c in n:
                        if c.tagname == "caption":
                            sectname = c.astext()
                            break

                    node["ids"].remove(labelid)
                    node["names"].remove(name)
                    n["ids"].append(labelid)
                    n["names"].append(name)
                    document.settings.env.labels[name] = (
                        document.settings.env.docname,
                        labelid,
                        sectname,
                    )
                    break


class ExecuteDirective(Directive):
    """The ``.. execute::`` directive, as documented in the module's docstring."""

    has_content = True
    required_arguments = 0
    optional_arguments = 2
    final_argument_whitespace = False
    option_spec = {
        "alt": directives.unchanged,
        "height": directives.length_or_unitless,
        "width": directives.length_or_percentage_or_unitless,
        "scale": directives.nonnegative_int,
        "align": Image.align,
        "class": directives.class_option,
        "include-source": _option_boolean,
        "show-return": _option_boolean,
        "show-output": _option_output_format,
        "show-output-label": _option_boolean,
        "show-return-label": _option_boolean,
        "show-source-link": _option_boolean,
        "card": _option_boolean,
        "carousel": _option_carousel,
        "card-width": _option_card_width,
        "format": _option_format,
        "context": _option_context,
        "nofigs": directives.flag,
        "caption": directives.unchanged,
        "link-text": directives.unchanged,
    }

    def run(self):
        """Run the plot directive."""
        try:
            rcParams = {
                "image.cmap": "wildboar",
                "axes.prop_cycle": cycler(color=colors),
            }
            light_dark = (
                self.state_machine.document.settings.env.config.execute_light_dark
            )

            if light_dark:
                light, is_doctest, build_dir_link, source_file_name, errors_light = run(
                    arguments=self.arguments,
                    content=self.content,
                    options=self.options,
                    rc_params=rcParams,
                    dark_mode=False,
                    state_machine=self.state_machine,
                    state=self.state,
                    lineno=self.lineno,
                )
                if is_doctest:
                    raise ValueError("doctest syntax is not supported")

                dark, _, _, _, errors_dark = run(
                    arguments=self.arguments,
                    content=self.content,
                    options=self.options,
                    rc_params=rcParams,
                    dark_mode=True,
                    state_machine=self.state_machine,
                    state=self.state,
                    lineno=self.lineno,
                )

                if "align" not in self.options:
                    self.options["align"] = "center"

                opts = [
                    f":{key}: {val}"
                    for key, val in self.options.items()
                    if key in ("alt", "height", "width", "scale", "align")
                ]
                light_opts = opts + [":class: only-light, card-image"]
                dark_opts = opts + [":class: only-dark, card-image"]
                if "card" in self.options and not self.options["card"]:
                    template = TEMPLATE
                else:
                    template = CARD_TEMPLATE

                total_lines = []
                for (source_code, (kind, output), stdout, src_name, light_images), (
                    _,
                    _,
                    _,
                    _,
                    dark_images,
                ) in zip(light, dark):
                    result = jinja2.Template(template).render(
                        default_fmt="png",
                        images=list(zip(light_images, dark_images)),
                        build_dir=build_dir_link,
                        source_code=source_code,
                        src_name=src_name,
                        carousel=self.options.get("carousel", 2),
                        card_width=self.options.get("card-width", "auto"),
                        show_return=self.options.get("show-return", False),
                        show_return_label=self.options.get("show-return-label", True),
                        show_output_label=self.options.get("show-output-label", False),
                        output_format=self.options.get("show-output", None),
                        return_output=output,
                        return_kind=kind,
                        stdout=stdout,
                        link_text=self.options.get("link-text", "Download source"),
                        light_options=light_opts,
                        dark_options=dark_opts,
                    )
                    total_lines.extend(result.split("\n"))
            else:
                light, is_doctest, build_dir_link, source_file_name, errors_light = run(
                    arguments=self.arguments,
                    content=self.content,
                    options=self.options,
                    rc_params=rcParams,
                    dark_mode=False,
                    state_machine=self.state_machine,
                    state=self.state,
                    lineno=self.lineno,
                )
                if is_doctest:
                    raise ValueError("doctest syntax is not supported")

                if "align" not in self.options:
                    self.options["align"] = "center"

                opts = [
                    f":{key}: {val}"
                    for key, val in self.options.items()
                    if key in ("alt", "height", "width", "scale", "align")
                ]
                light_opts = opts + [":class: card-image"]
                dark_opts = None
                if "card" in self.options and not self.options["card"]:
                    template = TEMPLATE
                else:
                    template = CARD_TEMPLATE

                total_lines = []
                for source_code, (
                    kind,
                    output,
                ), stdout, src_name, light_images in light:
                    result = jinja2.Template(template).render(
                        default_fmt="png",
                        images=list(zip(light_images, light_images)),
                        build_dir=build_dir_link,
                        source_code=source_code,
                        src_name=src_name,
                        carousel=self.options.get("carousel", 2),
                        card_width=self.options.get("card-width", "auto"),
                        show_return=self.options.get("show-return", False),
                        show_return_label=self.options.get("show-return-label", True),
                        show_output_label=self.options.get("show-output-label", False),
                        output_format=self.options.get("show-output", None),
                        return_output=output,
                        return_kind=kind,
                        stdout=stdout,
                        link_text=self.options.get("link-text", "Download source"),
                        light_options=light_opts,
                        dark_options=dark_opts,
                    )
                    total_lines.extend(result.split("\n"))

            if total_lines:
                self.state_machine.insert_input(total_lines, source=source_file_name)

            return errors_light
        except Exception as e:
            raise self.error(str(e))


def _copy_css_file(app, exc):
    if exc is None and app.builder.format == "html":
        src = cbook._get_data_path("plot_directive/plot_directive.css")
        dst = app.outdir / Path("_static")
        dst.mkdir(exist_ok=True)
        # Use copyfile because we do not want to copy src's permissions.
        shutil.copyfile(src, dst / Path("plot_directive.css"))


# -----------------------------------------------------------------------------
# Doctest handling
# -----------------------------------------------------------------------------


def contains_doctest(text):
    try:
        # check if it's valid Python as-is
        compile(text, "<string>", "exec")
        return False
    except SyntaxError:
        pass
    r = re.compile(r"^\s*>>>", re.M)
    m = r.search(text)
    return bool(m)


def _split_code_at_show(text, function_name):
    """Split code at plt.show()."""

    is_doctest = contains_doctest(text)
    if function_name is None:
        parts = []
        part = []
        for line in text.split("\n"):
            if (not is_doctest and line.startswith("plt.show(")) or (
                is_doctest and line.strip() == ">>> plt.show()"
            ):
                part.append(line)
                parts.append("\n".join(part))
                part = []
            else:
                part.append(line)
        if "\n".join(part).strip():
            parts.append("\n".join(part))
    else:
        parts = [text]
    return is_doctest, parts


# -----------------------------------------------------------------------------
# Template
# -----------------------------------------------------------------------------

_SOURCECODE = """
{{ source_code }}
"""

TEMPLATE_SRCSET = (
    _SOURCECODE
    + """
   {% for img in images %}
   .. figure-mpl:: {{ build_dir }}/{{ img.basename }}.{{ default_fmt }}
      {% for option in options -%}
      {{ option }}
      {% endfor %}
      {%- if caption -%}
      {{ caption }}  {# appropriate leading whitespace added beforehand #}
      {% endif -%}
      {%- if srcset -%}
        :srcset: {{ build_dir }}/{{ img.basename }}.{{ default_fmt }}
        {%- for sr in srcset -%}
            , {{ build_dir }}/{{ img.basename }}.{{ sr }}.{{ default_fmt }} {{sr}}
        {%- endfor -%}
      {% endif %}

   {% endfor %}

"""
)

TEMPLATE = """
{% if source_code %}
.. code-block:: python
   {{ source_code_class }}

{{ source_code }}
{% endif %}
{% for light, dark in images %}

.. image:: {{ build_dir }}/{{ light.basename }}.{{ default_fmt }}
   {% for option in light_options -%}
   {{ option }}
   {% endfor %}

.. image:: {{ build_dir }}/{{ dark.basename }}.{{ default_fmt }}
   {% for option in dark_options -%}
   {{ option }}
   {% endfor %}
{% endfor %}
{%- if src_name -%}
:download:`Source code <{{ build_dir }}/{{ src_name }}>`
{%- endif -%}
"""

CARD_TEMPLATE = """
{%- if images or source_code or (show_return and return_kind) or stdout %}
.. grid:: 1
    :padding: 0
    :margin: 0
    {% if source_code or (show_return and return_kind) or stdout %}
    .. grid-item::
        :columns: 12
    {% if source_code %}
        .. code-block:: python
{% filter indent(width=8) %}
{{ source_code }}
{%- endfilter %}
    {%- endif %}
    {% if show_return and return_kind %}
        {%- if return_kind == "html" %}
        .. raw:: html

            <div class="sd-text-center", data-mime-type="text/html">
{%- filter indent(width=12) %}
{{ return_output|trim }}
{%- endfilter %}
            </div>
        {%- elif return_kind == "plain" %}
        .. code-block::
            :class: sd-mt-0 {%- if show_return_label %} execute-return {%- endif %}
{% filter indent(width=12) %}
{{ return_output }}
{%- endfilter %}
        {%- endif %}
    {%- endif %}
    {% if output_format == "code" %}
        .. code-block::
            :class: sd-mt-0 {%- if show_output_label %} execute-output {%- endif %}

{% filter indent(width=12) %}
{{ stdout }}
{%- endfilter %}
    {%- endif %}
    {%- endif -%}
{% if output_format == "plain" %}
{{stdout|trim}}
{% if images %}
.. grid:: 1
    :padding: 0
    :margin: 0
{% endif -%}
{% endif -%}
    {% if images %}
    .. grid-item::
        :columns: 12
        {% if images|length > 0 %}
        .. card-carousel:: {{ carousel }}
        {% for light, dark in images %}
            .. card::
                :width: {{ card_width }}
                :shadow: none
                :text-align: center
                :margin: auto

                .. image:: {{ build_dir }}/{{ light.basename }}.{{ default_fmt }}
                    {% for option in light_options -%}
                    {{ option }}
                    {% endfor %}
                {% if light != dark -%}
                .. image:: {{ build_dir }}/{{ dark.basename }}.{{ default_fmt }}
                    {% for option in dark_options -%}
                    {{ option }}
                    {% endfor %}
                {% endif -%}

        {%- endfor %}
        {% else -%}
        {% for light, dark in images %}
        .. card::
            :width: {{ card_width }}
            :shadow: none
            :text-align: center
            :margin: auto

            .. image:: {{ build_dir }}/{{ light.basename }}.{{ default_fmt }}
                {% for option in light_options -%}
                {{ option }}
                {% endfor %}

            .. image:: {{ build_dir }}/{{ dark.basename }}.{{ default_fmt }}
                {% for option in dark_options -%}
                {{ option }}
                {% endfor %}

        {%- endfor %}
        {%- endif %}
    {%- if src_name %}
    .. grid-item::
        :columns: 12
        :child-align: center
        :class: sd-text-center

        :download:`{{ link_text }} <{{ build_dir }}/{{ src_name }}>`
    {%- endif %}
    {%- endif %}
{%- elif src_name -%}
:download:`{{ link_text }} <{{ build_dir }}/{{ src_name }}>`
{%- endif %}

"""
exception_template = """
.. only:: html

   [`source code <%(linkdir)s/%(basename)s.py>`__]

Exception occurred rendering plot.

"""

# the context of the plot for all directives specified with the
# :context: option
plot_context = dict()


class ImageFile:
    def __init__(self, basename, dirname):
        self.basename = basename
        self.dirname = dirname
        self.formats = []

    def filename(self, format):
        return os.path.join(self.dirname, f"{self.basename}.{format}")

    def filenames(self):
        return [self.filename(fmt) for fmt in self.formats]


def out_of_date(original, derived, includes=None):
    """
    Return whether *derived* is out-of-date relative to *original* or any of
    the RST files included in it using the RST include directive (*includes*).
    *derived* and *original* are full paths, and *includes* is optionally a
    list of full paths which may have been included in the *original*.
    """
    if not os.path.exists(derived):
        return True

    if includes is None:
        includes = []
    files_to_check = [original, *includes]

    def out_of_date_one(original, derived_mtime):
        return os.path.exists(original) and derived_mtime < os.stat(original).st_mtime

    derived_mtime = os.stat(derived).st_mtime
    return any(out_of_date_one(f, derived_mtime) for f in files_to_check)


class PlotError(RuntimeError):
    pass


def exec_with_return(code, globals=None, locals=None):
    a = ast.parse(code)
    last_expression = None
    if a.body:
        if isinstance(a_last := a.body[-1], ast.Expr):
            last_expression = ast.unparse(a.body.pop())
        elif isinstance(a_last, ast.Assign):
            last_expression = ast.unparse(a_last.targets[0])
        elif isinstance(a_last, (ast.AnnAssign, ast.AugAssign)):
            last_expression = ast.unparse(a_last.target)
    exec(ast.unparse(a), globals, locals)
    if last_expression:
        return eval(last_expression, globals, locals)


def _run_code(code, code_path, ns=None, function_name=None):
    """
    Import a Python module from a path, and run the function given by
    name, if function_name is not None.
    """

    # Change the working directory to the directory of the example, so
    # it can get at its data files, if any.  Add its path to sys.path
    # so it can import any helper modules sitting beside it.
    pwd = os.getcwd()
    if setup.config.execute_working_directory is not None:
        try:
            os.chdir(setup.config.execute_working_directory)
        except OSError as err:
            raise OSError(
                f"{err}\n`execute_working_directory` option in "
                f"Sphinx configuration file must be a valid "
                f"directory path"
            ) from err
        except TypeError as err:
            raise TypeError(
                f"{err}\n`execute_working_directory` option in "
                f"Sphinx configuration file must be a string or "
                f"None"
            ) from err
    elif code_path is not None:
        dirname = os.path.abspath(os.path.dirname(code_path))
        os.chdir(dirname)

    return_value = None
    stdout = None
    with StringIO() as buf, cbook._setattr_cm(
        sys, argv=[code_path], path=[os.getcwd(), *sys.path]
    ), contextlib.redirect_stdout(buf):
        try:
            if ns is None:
                ns = {}
            if not ns:
                if setup.config.execute_pre_code is None:
                    exec(
                        "import numpy as np\n" "from matplotlib import pyplot as plt\n",
                        ns,
                    )
                else:
                    exec(str(setup.config.execute_pre_code), ns)
            if "__main__" in code:
                ns["__name__"] = "__main__"

            # Patch out non-interactive show() to avoid triggering a warning.
            with cbook._setattr_cm(FigureManagerBase, show=lambda self: None):
                return_value = exec_with_return(code, ns)
                if function_name is not None:
                    return_value = eval(function_name + "()", ns)

            stdout = buf.getvalue()
        except (Exception, SystemExit) as err:
            raise PlotError(traceback.format_exc()) from err
        finally:
            os.chdir(pwd)

    return ns, return_value, stdout


def clear_state(execute_rcparams, *, close=True):
    if close:
        plt.close("all")
    matplotlib.rcParams.update(execute_rcparams)


def get_plot_formats(config):
    default_dpi = {"png": 80, "hires.png": 200, "pdf": 200}
    formats = []
    execute_formats = config.execute_formats
    for fmt in execute_formats:
        if isinstance(fmt, str):
            if ":" in fmt:
                suffix, dpi = fmt.split(":")
                formats.append((str(suffix), int(dpi)))
            else:
                formats.append((fmt, default_dpi.get(fmt, 80)))
        elif isinstance(fmt, (tuple, list)) and len(fmt) == 2:
            formats.append((str(fmt[0]), int(fmt[1])))
        else:
            raise PlotError('invalid image format "%r" in execute_formats' % fmt)
    return formats


def _parse_srcset(entries):
    """
    Parse srcset for multiples...
    """
    srcset = {}
    for entry in entries:
        entry = entry.strip()
        if len(entry) >= 2:
            mult = entry[:-1]
            srcset[float(mult)] = entry
        else:
            raise ExtensionError(f"srcset argument {entry!r} is invalid.")
    return srcset


def render_figures(
    code,
    code_path,
    output_dir,
    output_base,
    context,
    function_name,
    config,
    dark_mode,
    rc_params={},
    context_reset=False,
    close_figs=False,
    code_includes=None,
):
    """
    Run a pyplot script and save the images in *output_dir*.

    Save the images under *output_dir* with file names derived from
    *output_base*
    """

    if function_name is not None:
        output_base = f"{output_base}_{function_name}"
    formats = get_plot_formats(config)

    # Try to determine if all images already exist

    is_doctest, code_pieces = _split_code_at_show(code, function_name)
    # Look for single-figure output files first
    img = ImageFile(output_base, output_dir)
    for format, dpi in formats:
        if context or out_of_date(
            code_path, img.filename(format), includes=code_includes
        ):
            all_exists = False
            break
        img.formats.append(format)
    else:
        all_exists = True

    if all_exists:
        return [(code, [img])]

    # Then look for multi-figure output files
    results = []
    for i, code_piece in enumerate(code_pieces):
        images = []
        for j in itertools.count():
            if len(code_pieces) > 1:
                img = ImageFile("%s_%02d_%02d" % (output_base, i, j), output_dir)
            else:
                img = ImageFile("%s_%02d" % (output_base, j), output_dir)
            for fmt, dpi in formats:
                if context or out_of_date(
                    code_path, img.filename(fmt), includes=code_includes
                ):
                    all_exists = False
                    break
                img.formats.append(fmt)

            # assume that if we have one, we have them all
            if not all_exists:
                all_exists = j > 0
                break
            images.append(img)
        if not all_exists:
            break
        results.append((code_piece, images))
    else:
        all_exists = True

    if all_exists:
        return results

    # We didn't find the files, so build them

    results = []
    ns = plot_context if context else {}

    if context_reset:
        clear_state(rc_params)
        plot_context.clear()

    close_figs = not context or close_figs

    for i, code_piece in enumerate(code_pieces):
        if dark_mode:
            matplotlib.style.use("dark_background")
        else:
            matplotlib.style.use("default")

        if not context or config.execute_apply_rcparams:
            clear_state(rc_params, close=close_figs)
        elif close_figs:
            plt.close("all")

        _ns, return_value, stdout = _run_code(
            doctest.script_from_examples(code_piece) if is_doctest else code_piece,
            code_path,
            ns,
            function_name,
        )

        return_repr = (None, None)
        if return_value is not None:
            if hasattr(return_value, "_repr_html_"):
                return_repr = ("html", return_value._repr_html_())
            else:
                return_repr = ("plain", return_value.__repr__())

        images = []
        fig_managers = _pylab_helpers.Gcf.get_all_fig_managers()
        for j, figman in enumerate(fig_managers):
            if len(fig_managers) == 1 and len(code_pieces) == 1:
                img = ImageFile(output_base, output_dir)
            elif len(code_pieces) == 1:
                img = ImageFile("%s_%02d" % (output_base, j), output_dir)
            else:
                img = ImageFile("%s_%02d_%02d" % (output_base, i, j), output_dir)
            images.append(img)

            for fmt, dpi in formats:
                try:
                    figman.canvas.figure.savefig(
                        img.filename(fmt), dpi=dpi, transparent=True
                    )
                    if fmt == formats[0][0] and config.execute_srcset:
                        # save a 2x, 3x etc version of the default...
                        srcset = _parse_srcset(config.execute_srcset)
                        for mult, suffix in srcset.items():
                            fm = f"{suffix}.{fmt}"
                            img.formats.append(fm)
                            figman.canvas.figure.savefig(
                                img.filename(fm), dpi=int(dpi * mult)
                            )
                except Exception as err:
                    raise PlotError(traceback.format_exc()) from err
                img.formats.append(fmt)

        results.append((code_piece, return_repr, stdout, images))

    if not context or config.execute_apply_rcparams:
        clear_state(rc_params, close=not context)

    return results


def run(
    *, arguments, content, options, rc_params, dark_mode, state_machine, state, lineno
):
    document = state_machine.document
    config = document.settings.env.config
    nofigs = "nofigs" in options

    if config.execute_srcset and setup.app.builder.name == "singlehtml":
        raise ExtensionError(
            "execute_srcset option not compatible with single HTML writer"
        )

    formats = get_plot_formats(config)
    default_fmt = formats[0][0]

    options.setdefault("include-source", config.execute_include_source)
    options.setdefault("show-source-link", config.execute_html_show_source_link)

    if "class" in options:
        # classes are parsed into a list of string, and output by simply
        # printing the list, abusing the fact that RST guarantees to strip
        # non-conforming characters
        options["class"] = ["plot-directive"] + options["class"]
    else:
        options.setdefault("class", ["plot-directive"])
    keep_context = "context" in options
    context_opt = None if not keep_context else options["context"]
    if context_opt is None:
        context_opt = config.execute_context

    rst_file = document.attributes["source"]
    rst_dir = os.path.dirname(rst_file)

    if len(arguments):
        if not config.execute_basedir:
            source_file_name = os.path.join(
                setup.app.builder.srcdir, directives.uri(arguments[0])
            )
        else:
            source_file_name = os.path.join(
                setup.confdir, config.execute_basedir, directives.uri(arguments[0])
            )
        # If there is content, it will be passed as a caption.
        caption = "\n".join(content)

        # Enforce unambiguous use of captions.
        if "caption" in options:
            if caption:
                raise ValueError(
                    "Caption specified in both content and options."
                    " Please remove ambiguity."
                )
            # Use caption option
            caption = options["caption"]

        # If the optional function name is provided, use it
        if len(arguments) == 2:
            function_name = arguments[1]
        else:
            function_name = None

        code = Path(source_file_name).read_text(encoding="utf-8")
        output_base = os.path.basename(source_file_name)
    else:
        source_file_name = rst_file
        code = textwrap.dedent("\n".join(map(str, content)))
        counter = document.attributes.get("_plot_counter", 0) + 1
        document.attributes["_plot_counter"] = counter
        base, ext = os.path.splitext(os.path.basename(source_file_name))
        output_base = "%s-%d.py" % (base, counter)
        function_name = None
        caption = options.get("caption", "")

    base, source_ext = os.path.splitext(output_base)
    if source_ext in (".py", ".rst", ".txt"):
        output_base = base
    else:
        source_ext = ""

    # ensure that LaTeX includegraphics doesn't choke in foo.bar.pdf filenames
    output_base = output_base.replace(".", "-")

    # is it in doctest format?
    is_doctest = contains_doctest(code)
    if "format" in options:
        if options["format"] == "python":
            is_doctest = False
        else:
            is_doctest = True

    # determine output directory name fragment
    source_rel_name = relpath(source_file_name, setup.confdir)
    source_rel_dir = os.path.dirname(source_rel_name).lstrip(os.path.sep)

    # build_dir: where to place output files (temporarily)
    build_dir = os.path.join(
        os.path.dirname(setup.app.doctreedir), "plot_directive", source_rel_dir
    )
    # get rid of .. in paths, also changes pathsep
    # see note in Python docs for warning about symbolic links on Windows.
    # need to compare source and dest paths at end
    build_dir = os.path.normpath(build_dir)
    os.makedirs(build_dir, exist_ok=True)

    # how to link to files from the RST file
    try:
        build_dir_link = relpath(build_dir, rst_dir).replace(os.path.sep, "/")
    except ValueError:
        # on Windows, relpath raises ValueError when path and start are on
        # different mounts/drives
        build_dir_link = build_dir

    # get list of included rst files so that the output is updated when any
    # plots in the included files change. These attributes are modified by the
    # include directive (see the docutils.parsers.rst.directives.misc module).
    try:
        source_file_includes = [
            os.path.join(os.getcwd(), t[0]) for t in state.document.include_log
        ]
    except AttributeError:
        # the document.include_log attribute only exists in docutils >=0.17,
        # before that we need to inspect the state machine
        possible_sources = {
            os.path.join(setup.confdir, t[0]) for t in state_machine.input_lines.items
        }
        source_file_includes = [f for f in possible_sources if os.path.isfile(f)]
    # remove the source file itself from the includes
    try:
        source_file_includes.remove(source_file_name)
    except ValueError:
        pass

    # save script (if necessary)
    if options["show-source-link"]:
        Path(build_dir, output_base + source_ext).write_text(
            doctest.script_from_examples(code)
            if source_file_name == rst_file and is_doctest
            else code,
            encoding="utf-8",
        )

    # make figures
    try:
        results = render_figures(
            code=code,
            code_path=source_file_name,
            output_dir=build_dir,
            output_base=output_base,
            context=keep_context,
            function_name=function_name,
            config=config,
            rc_params=rc_params,
            dark_mode=dark_mode,
            context_reset=context_opt == "reset",
            close_figs=context_opt == "close-figs",
            code_includes=source_file_includes,
        )
        errors = []
    except PlotError as err:
        reporter = state.memo.reporter
        sm = reporter.system_message(
            2,
            "Exception occurred in plotting {}\n from {}:\n{}".format(
                output_base, source_file_name, err
            ),
            line=lineno,
        )
        results = [(code, None, None, [])]
        errors = [sm]

    # Properly indent the caption
    if caption and config.execute_srcset:
        caption = f":caption: {caption}"
    elif caption:
        caption = "\n" + "\n".join(
            "      " + line.strip() for line in caption.split("\n")
        )
    # generate output restructuredtext
    total_lines = []
    all_images = []
    for j, (code_piece, return_value, stdout, images) in enumerate(results):
        if options["include-source"]:
            if is_doctest:
                lines = code_piece.splitlines()
            else:
                lines = textwrap.indent(code_piece, "    ").splitlines()

            source_code = "\n".join(lines)
        else:
            source_code = None

        if j == 0 and options["show-source-link"]:
            src_name = output_base + source_ext
        else:
            src_name = None

        if nofigs:
            images = []

        all_images.append((source_code, return_value, stdout, src_name, images))

    return all_images, is_doctest, build_dir_link, source_file_name, errors


def setup(app):
    setup.app = app
    setup.config = app.config
    setup.confdir = app.confdir
    app.add_directive("execute", ExecuteDirective)
    app.add_config_value("execute_light_dark", True, True)
    app.add_config_value("execute_pre_code", None, True)
    app.add_config_value("execute_include_source", True, True)
    app.add_config_value("execute_html_show_source_link", False, True)
    app.add_config_value("execute_formats", ["png"], True)
    app.add_config_value("execute_basedir", None, True)
    app.add_config_value("execute_html_show_formats", True, True)
    app.add_config_value("execute_rcparams", {}, True)
    app.add_config_value("execute_apply_rcparams", True, True)
    app.add_config_value("execute_working_directory", None, True)
    app.add_config_value("execute_template", None, True)
    app.add_config_value("execute_srcset", [], True)
    app.add_config_value("execute_context", "close-figs", True)
    app.connect("doctree-read", mark_plot_labels)
    app.connect("build-finished", _copy_css_file)
    return {
        "version": "0.1",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



# Authors: Isak Samsten
# License: BSD 3 clause

import numpy as np
from sklearn.utils import column_or_1d, resample

from ..utils.validation import check_array

try:
    import matplotlib.pylab as plt
    from matplotlib.cm import get_cmap
    from matplotlib.collections import LineCollection
    from matplotlib.colors import Normalize
    from matplotlib.lines import Line2D
except ModuleNotFoundError as e:
    from ..utils import DependencyMissing

    matplotlib_missing = DependencyMissing(e, package="matplotlib")
    plt = matplotlib_missing
    get_cmap = matplotlib_missing
    Normalize = matplotlib_missing
    LineCollection = matplotlib_missing
    Line2D = matplotlib_missing


# noqa: H0002
class MidpointNormalize(Normalize):
    """
    Normalise values with a specified midpoint.

    Parameters
    ----------
    vmin : float, optional
        The minumum value.
    vmax : float, optional
        The maximum value.
    midpoint : float, optional
        The midpoint of the values.
    """

    def __init__(self, vmin=None, vmax=None, midpoint=None):
        super().__init__(vmin, vmax, clip=False)
        self.midpoint = midpoint

    def __call__(self, value, clip=None):
        x, y = [self.vmin, self.midpoint, self.vmax], [0, 0.5, 1]
        return np.ma.masked_array(np.interp(value, x, y), np.isnan(value))


def plot_time_domain(
    x,
    y=None,
    *,
    n_samples=100,
    ax=None,
    alpha=0.5,
    linewidth=0.5,
    zorder=-1,
    cmap="Dark2",
    show_legend=8,
):
    """
    Plot the samples in the time domain.

    Parameters
    ----------
    x : array-like of shape (n_sample, n_timestep)
        The samples.
    y : array-like of shape (n_samples, ), optional
        The labels, assumed to be discrete labels.
    n_samples : int, optional
        The maximum number of samples to plot. If n_samples is larger than the number
        of samples in x or None, all samples are plotted.
    ax : Axes, optional
        The matplotlib Axes-object.
    alpha : float, optional
        The opacity of the samples.
    linewidth : float, optional
        The width of the sample lines.
    zorder : int, optional
        The order where the samples are plotted. By default we plot the samples
        at -1.
    cmap : str, optional
        The colormap used to colorize samples according to its label.
    show_legend : bool or int, optional
        Whether the legend of labels are show.

        - if bool, show the legend if y is not None
        - if int, show the legend if the number of labels are lower than the
          show_legend parameter value

    Returns
    -------
    Axes
        The axes object that has been plotted.

    Examples
    --------
    >>> from wildboar.utils.plot import plot_time_domain
    >>> from wildboar.datasets import load_gun_point
    >>> X, y = load_gun_point(X, y)
    >>> plot_time_domain(X, y, n_sample=10)
    """
    if ax is None:
        fig, ax = plt.subplots()

    x = check_array(np.atleast_2d(x), input_name="x", allow_3d=False, order=None)
    if y is not None:
        y = column_or_1d(y, warn=True)
        if y.shape[0] != x.shape[0]:
            raise ValueError("The number of labels and samples are not the same.")

        labels, idx, inv = np.unique(y, return_inverse=True, return_index=True)
        cmap = get_cmap(cmap, len(labels))
    else:
        cmap = get_cmap(cmap, 1)
        inv = np.zeros(x.shape[0])

    if n_samples is not None:
        x, inv = resample(
            x,
            inv,
            replace=False,
            n_samples=n_samples if n_samples < x.shape[0] else x.shape[0],
            stratify=y,
            random_state=0,
        )

    x_axis = np.arange(x.shape[-1] + 1)
    collection = LineCollection(
        [list(zip(x_axis, x[i])) for i in range(x.shape[0])],
        colors=[cmap(inv[i]) for i in range(x.shape[0])],
        zorder=zorder,
        linewidth=linewidth,
        alpha=alpha,
    )
    ax.add_collection(collection)

    if y is not None and (show_legend is True or len(labels) <= show_legend):
        legend = ax.legend(
            [Line2D([0], [0], color=cmap(i)) for i in idx],
            labels,
            loc="best",
            ncol=(len(labels) // 3) + 1,
        )
        legend.set_zorder(100)
        ax.add_artist(legend)

    ax.set_xlim([1, x.shape[-1] - 1])
    ax.set_ylim([np.min(x) - np.std(x), np.max(x) + np.std(x)])

    return ax


def plot_frequency_domain(
    x,
    y=None,
    *,
    ax=None,
    n_samples=100,
    jitter=False,
    sample_spacing=1,
    frequency=False,
    cmap="Dark2",
):
    """
    Plot the samples in the freqency domain.

    Parameters
    ----------
    x : array-like of shape (n_sample, n_timestep)
        The samples.
    y : array-like of shape (n_samples, ), optional
        The labels.
    ax : Axes, optional
        The matplotlib Axes-object.
    n_samples : int, optional
        The maximum number of samples to plot. If n_samples is larger than the number
        of samples in x or None, all samples are plotted.
    jitter : bool, optional
        Add jitter to the amplitude lines.
    sample_spacing : int, optional
        The frequency domain sample spacing.
    frequency : bool, optional
        Show the frequency bins.
    cmap : str, optional
        The colormap.

    Returns
    -------
    Axes
        The axes object that has been plotted.

    Examples
    --------
    >>> from wildboar.utils.plot import plot_frequency_domain
    >>> from wildboar.datasets import load_gun_point
    >>> X, y = load_gun_point(X, y)
    >>> plot_frequency_domain(X, y, n_sample=10)
    """
    if ax is None:
        fig, ax = plt.subplots()

    x = check_array(x, allow_3d=False, order=None)
    if y is not None:
        y = check_array(y, ensure_2d=False, allow_nd=False, order=None)
        if y.shape[0] != x.shape[0]:
            raise ValueError("The number of samples and lables are not the same.")

        labels, idx, inv = np.unique(y, return_inverse=True, return_index=True)
        cmap = get_cmap(cmap, len(labels))
    else:
        cmap = get_cmap(cmap, 1)
        inv = np.zeros(x.shape[0])

    x, inv = resample(
        x,
        inv,
        replace=False,
        n_samples=n_samples if n_samples < x.shape[0] else x.shape[0],
        stratify=y,
        random_state=0,
    )

    n_freqs = int(x.shape[-1] // 2)
    x_freq = np.abs(np.fft.fft(x, axis=1)[:, 1 : n_freqs + 1])  # / n_freqs
    x_axis = np.arange(1, n_freqs + 1)
    max_freq = np.max(x_freq)
    if frequency:
        for i in x_axis:
            ax.axvspan(
                i - 0.5,
                i + 0.5,
                0,
                1,
                facecolor=None,
                edgecolor="gray",
                fill=False,
                alpha=0.05,
                zorder=-100,
            )
    ax.set_ylabel("Amplitude")
    for i in range(x.shape[0]):
        if jitter:
            x_axis_tmp = x_axis + np.random.normal(scale=0.5, size=n_freqs)
        else:
            x_axis_tmp = x_axis

        ax.vlines(
            x_axis_tmp,
            0,
            x_freq[i],
            color=cmap(inv[i]),
            alpha=0.3,
            linewidth=1,
            zorder=-1,
        )

    if y is not None:
        legend = ax.legend(
            [Line2D([0], [0], color=cmap(i)) for i in idx],
            labels,
            loc="best",
            ncol=(len(labels) // 3) + 1,
        )

        ax.add_artist(legend)

    # ticks = ax.get_xticks().astype(int)[: len(x_axis)]
    # ticks[0] = 1
    # ax.set_xticks(ticks)
    # x_label = np.fft.fftfreq(x.shape[-1], d=sample_spacing)[ticks]
    # ax.set_xticklabels("%.2f" % lab for lab in x_label)
    ax.set_xlim([0.5, n_freqs + 0.5])
    ax.set_ylim(0, max_freq + np.std(x_freq))

    return ax



# cython: language_level=3
# cython: infer_types=True
# cython: boundscheck=False
# Fast swap step and build step in PAM algorithm for k_medoid.
# Author: Timothée Mathieu
# License: 3-clause BSD
# Adapted from: https://github.com/scikit-learn-contrib/scikit-learn-extra
# /blob/main/sklearn_extra/cluster/_k_medoids_helper.pyx#L72

import numpy as np


def _pam_optimal_swap(
    double[:, :] D,
    Py_ssize_t[:] medoid_idxs,
    Py_ssize_t[:] not_medoid_idxs,
    double[:] Djs,
    double[:] Ejs,
    Py_ssize_t n_clusters
):
    # Initialize best cost change and the associated swap couple.
    cdef (Py_ssize_t, Py_ssize_t, double) best_cost_change = (1, 1, 0.0)
    cdef Py_ssize_t sample_size = D.shape[0]
    cdef Py_ssize_t i, j, h, id_i, id_h, id_j
    cdef double cost_change
    cdef Py_ssize_t not_medoid_shape = sample_size - n_clusters
    cdef bint cluster_i_bool, not_cluster_i_bool, second_best_medoid
    cdef bint not_second_best_medoid

    # Compute the change in cost for each swap.
    for h in range(not_medoid_shape):
        # id of the potential new medoid.
        id_h = not_medoid_idxs[h]
        for i in range(n_clusters):
            # id of the medoid we want to replace.
            id_i = medoid_idxs[i]
            cost_change = 0.0
            # compute for all not-selected points the change in cost
            for j in range(not_medoid_shape):
                id_j = not_medoid_idxs[j]
                cluster_i_bool = D[id_i, id_j] == Djs[id_j]
                not_cluster_i_bool = D[id_i, id_j] != Djs[id_j]
                second_best_medoid = D[id_h, id_j] < Ejs[id_j]
                not_second_best_medoid = D[id_h, id_j] >= Ejs[id_j]

                if cluster_i_bool & second_best_medoid:
                    cost_change += D[id_j, id_h] - Djs[id_j]
                elif cluster_i_bool & not_second_best_medoid:
                    cost_change += Ejs[id_j] - Djs[id_j]
                elif not_cluster_i_bool & (D[id_j, id_h] < Djs[id_j]):
                    cost_change += D[id_j, id_h] - Djs[id_j]

            # same for i
            second_best_medoid = D[id_h, id_i] < Ejs[id_i]
            if  second_best_medoid:
                cost_change += D[id_i, id_h]
            else:
                cost_change += Ejs[id_i]

            if cost_change < best_cost_change[2]:
                best_cost_change = (id_i, id_h, cost_change)

    # If one of the swap decrease the objective, return that swap.
    if best_cost_change[2] < 0:
        return best_cost_change
    else:
        return None


def _pam_build(double[:, :] D, Py_ssize_t n_clusters):
    """Compute BUILD initialization, a greedy medoid initialization."""

    cdef Py_ssize_t[:] medoid_idxs = np.zeros(n_clusters, dtype=int)
    cdef Py_ssize_t sample_size = len(D)
    cdef Py_ssize_t[:] not_medoid_idxs = np.arange(sample_size, dtype=int)
    cdef Py_ssize_t i, j,  id_i, id_j

    medoid_idxs[0] = np.argmin(np.sum(D, axis=0))
    not_medoid_idxs = np.delete(not_medoid_idxs, medoid_idxs[0])

    cdef Py_ssize_t n_medoids_current = 1

    cdef double[:] Dj = D[medoid_idxs[0]].copy()
    cdef double cost_change
    cdef (Py_ssize_t, Py_ssize_t) new_medoid = (0, 0)
    cdef double cost_change_max

    for _ in range(n_clusters -1):
        cost_change_max = 0

        for i in range(sample_size - n_medoids_current):
            id_i = not_medoid_idxs[i]
            cost_change = 0

            for j in range(sample_size - n_medoids_current):
                id_j = not_medoid_idxs[j]
                cost_change += max(0, Dj[id_j] - D[id_i, id_j])

            if cost_change >= cost_change_max:
                cost_change_max = cost_change
                new_medoid = (id_i, i)

        medoid_idxs[n_medoids_current] = new_medoid[0]
        n_medoids_current += 1
        not_medoid_idxs = np.delete(not_medoid_idxs, new_medoid[1])

        for id_j in range(sample_size):
            Dj[id_j] = min(Dj[id_j], D[id_j, new_medoid[0]])
    return medoid_idxs.base



# Authors: Isak Samsten
# License: BSD 3 clause

import math
import warnings
from abc import ABCMeta

import numpy as np
from sklearn.model_selection._split import _build_repr
from sklearn.utils import check_random_state
from sklearn.utils.validation import check_array


class RepeatedOutlierSplit(metaclass=ABCMeta):
    """
    Repeated random outlier cross-validator.

    Parameters
    ----------
    n_splits : int, optional
        The maximum number of splits.
        - if None, the number of splits is determined by the number of outliers
        as, `total_n_outliers/(n_inliers * n_outliers)`
        - if int, the number of splits is an upper-bound.
    test_size : float, optional
        The size of the test set.
    n_outlier : float, optional
        The fraction of outliers in the training and test sets.
    shuffle : bool, optional
        Shuffle the training indicies in each iteration.
    random_state : int or RandomState, optional
        The psudo-random number generator.

    Notes
    -----
    Contrary to other cross-validation strategies, the random outlier
    cross-validator does not ensure that all folds will be different.
    Instead, the inlier samples are shuffled and new outlier samples
    are inserted in the training and test sets repeatedly.
    """

    def __init__(
        self,
        n_splits=None,
        *,
        test_size=0.2,
        n_outlier=0.05,
        shuffle=True,
        random_state=None,
    ):
        self.n_outliers = n_outlier
        self.test_size = test_size
        self.n_splits = n_splits
        self.shuffle = shuffle
        self.random_state = random_state

    def get_n_splits(self, X, y, groups=None):
        """
        Return the number of splitting iterations in the cross-validator.

        Parameters
        ----------
        X : object
            The samples.
        y : object
            The labels.
        groups : object, optional
            Always ignored, exists for compatibility.

        Returns
        -------
        int
            Returns the number of splitting iterations in the cross-validator.
        """
        outlier_index = (y == -1).nonzero()[0]
        inlier_index = (y == 1).nonzero()[0]
        total_n_outliers = len(outlier_index)
        n_outliers = min(
            total_n_outliers, math.ceil(inlier_index.shape[0] * self.n_outliers)
        )
        if self.n_splits is None:
            n_splits = total_n_outliers // n_outliers
        else:
            n_splits = min(self.n_splits, total_n_outliers // n_outliers)
        return n_splits

    def split(self, x, y, groups=None):
        """
        Return training and test indicies.

        Parameters
        ----------
        x : object
            Always ignored, exists for compatibility.
        y : object
            The labels.
        groups : object, optional
            Always ignored, exists for compatibility.

        Yields
        ------
        train_idx, test_idx : ndarray
            The training and test indicies
        """
        y = check_array(y, ensure_2d=False)
        random_state = check_random_state(self.random_state)
        outlier_index = (y == -1).nonzero()[0]
        inlier_index = (y == 1).nonzero()[0]
        total_n_outliers = len(outlier_index)
        n_inliers = inlier_index.shape[0]
        n_outliers = min(
            total_n_outliers, math.ceil(inlier_index.shape[0] * self.n_outliers)
        )
        if n_outliers < 2:
            warnings.warn("to few outliers", UserWarning)

        if self.n_splits is None:
            n_splits = total_n_outliers // n_outliers
        else:
            n_splits = min(self.n_splits, total_n_outliers // n_outliers)
        random_state.shuffle(outlier_index)

        for i in range(n_splits):
            if self.shuffle:
                random_state.shuffle(inlier_index)
            outlier_index_sample = outlier_index[
                (i * n_outliers) : (i * n_outliers + n_outliers)
            ]
            outlier_test_index = math.ceil(n_outliers * self.test_size)
            inlier_test_index = math.ceil(n_inliers * self.test_size)

            yield (
                np.concatenate(
                    [
                        inlier_index[inlier_test_index:],  # last elements
                        outlier_index_sample[outlier_test_index:],
                    ]
                ),
                np.concatenate(
                    [
                        inlier_index[:inlier_test_index],  # first elements
                        outlier_index_sample[:outlier_test_index],
                    ]
                ),
            )

    def __repr__(self):
        return _build_repr(self)



import numbers

import numpy as np
from sklearn.base import ClassifierMixin, _fit_context
from sklearn.model_selection import GridSearchCV, LeaveOneOut
from sklearn.utils._param_validation import Interval, StrOptions
from sklearn.utils.validation import check_is_fitted

from ..base import BaseEstimator
from ..distance._distance import _METRICS
from ..distance._multi_metric import make_parameter_grid
from ..distance._neighbors import KNeighborsClassifier
from ..utils.validation import check_classification_targets


def _make_elastic_parameter_grid(std):
    return {
        "dtw": {"min_r": 0.01, "max_r": 0.3, "num_r": 10},
        "adtw": {
            "min_r": 0.01,
            "max_r": 0.3,
            "num_r": 3,
            "min_p": 1,
            "max_p": 4,
            "num_p": 3,
        },
        "ddtw": {"min_r": 0.01, "max_r": 0.3, "num_r": 10},
        "wdtw": {"min_g": 0.01, "max_g": 0.5, "num_g": 10},
        "wddtw": {"min_g": 0.01, "max_g": 0.5, "num_g": 10},
        "lcss": {
            "min_r": 0.0,
            "max_r": 0.25,
            "num_r": 3,
            "min_epsilon": 0.2 * std,
            "max_epsilon": std,
            "num_epsilon": 3,
        },
        "erp": {"min_g": 0, "max_g": 1.0, "num_g": 10},
        "msm": {"min_c": 0.01, "max_c": 100, "num_c": 10},
        "twe": {
            "min_penalty": 1e-5,
            "max_penalty": 1.0,
            "num_penalty": 3,
            "min_stiffness": 1e-6,
            "max_stiffness": 0.1,
            "num_stiffness": 3,
        },
    }


def _make_non_elastic_parameter_grid():
    return {
        "euclidean": None,
        "minkowski": {"min_p": -5, "max_p": 5, "num_p": 10},
        "chebyshev": None,
        "manhattan": None,
        "angular": None,
        "cosine": None,
    }


class ElasticEnsembleClassifier(ClassifierMixin, BaseEstimator):
    """
    Ensemble of :class:`wildboar.distance.KNeighborsClassifier`.

    Each classifier is fitted with an optimized parameter grid
    over metric parameters.

    Parameters
    ----------
    n_neighbors : int, optional
        The number of neighbors.
    metric : {"auto", "elastic", "non_elastic", "all"} or dict, optional
        The metric specification.

        - if "auto" or "elastic", fit one classifier for each elastic distance
          as described by Lines and Bagnall (2015). We use a slightly smaller
          parameter grid.
        - if "non_elastic", fit one classifier for each non-elastic distance
          measure.
        - if "all", fit one classifier for the metrics in both "elastic" and
          "non_elastic".
        - if dict, a custom metric specification.
    n_jobs : int, optional
        The number of paralell jobs.

    Attributes
    ----------
    scores : tuple
        A tuple of metric name and cross-validation score.

    References
    ----------
    Jason Lines and Anthony Bagnall,
        Time Series Classification with Ensembles of Elastic Distance Measures,
        Data Mining and Knowledge Discovery, 29(3), 2015.

    Examples
    --------
    >>> from wildboar.datasets import load_gun_point
    >>> from wildboar.ensemble import ElasticEnsembleClassifier
    >>> X_train, X_test, y_train, y_test = load_gun_point(merge_train_test=False)
    >>> clf = ElasticEnsembleClassifier(
    ...     metric={
    ...         "dtw": {"min_r": 0.1, "max_r": 0.3},
    ...         "ddtw": {"min_r": 0.1, "max_r": 0.3},
    ...     },
    ... )
    >>> clf.fit(X_train, y_train)
    ElasticEnsembleClassifier(metric={'ddtw': {'max_r': 0.3, 'min_r': 0.1},
                                      'dtw': {'max_r': 0.3, 'min_r': 0.1}})
    >>> clf.score(X_test, y_test)
    0.9866666666666667
    """

    _parameter_constraints = {
        "n_neighbors": [Interval(numbers.Integral, 1, None, closed="left")],
        "metric": [dict, StrOptions({"all", "non_elastic", "elastic", "auto"})],
        "n_jobs": [int, None],
    }

    def __init__(self, n_neighbors=1, *, metric="auto", n_jobs=None):
        self.n_neighbors = n_neighbors
        self.metric = metric
        self.n_jobs = n_jobs

    @_fit_context(prefer_skip_nested_validation=True)
    def fit(self, x, y):
        """
        Fit the estimator.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_timesteps) or (n_samples, n_dim, n_timesteps)
            The input samples.
        y : array-like of shape (n_samples, )
            The input labels.

        Returns
        -------
        object
            This estimator.
        """
        x, y = self._validate_data(x, y, allow_3d=True)
        check_classification_targets(y)

        self.classes_ = np.unique(y)
        if len(self.classes_) < 2:
            raise ValueError("too few labels")

        if self.metric in ["elastic", "auto"]:
            metric = _make_elastic_parameter_grid(x.std())
        elif self.metric == "non_elastic":
            metric = _make_non_elastic_parameter_grid()
        elif self.metric == "all":
            metric = {
                **_make_non_elastic_parameter_grid(),
                **_make_elastic_parameter_grid(x.std()),
            }
        else:
            metric = self.metric

        metric_param_grid = {}
        for metric_name, param_grid in metric.items():
            if metric_name not in _METRICS:
                raise ValueError(f"{metric_name} is not supported")
            metric_param_grid[metric_name] = make_parameter_grid(param_grid)

        self.estimators_ = []
        self.scores_ = []
        for metric, metric_params in metric_param_grid.items():
            gridcv = GridSearchCV(
                KNeighborsClassifier(n_neighbors=self.n_neighbors, metric=metric),
                param_grid={"metric_params": metric_params},
                cv=LeaveOneOut(),
                n_jobs=self.n_jobs,
            ).fit(x, y)
            estimator = gridcv.best_estimator_.set_params(n_jobs=self.n_jobs).fit(x, y)
            self.estimators_.append(estimator)
            self.scores_.append((metric, gridcv.best_score_))

        return self

    def predict_proba(self, x):
        """
        Compute probability estimates for the samples in x.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_timesteps) or (n_samples, n_dim, n_timesteps)
            The input time series.

        Returns
        -------
        ndarray of shape (n_samples, n_classes)
            The probabilities.
        """
        check_is_fitted(self)
        x = self._validate_data(x, allow_3d=True, reset=False)
        proba = np.zeros((x.shape[0], len(self.estimators_), len(self.classes_)))
        score_sum = 0
        for i, ((_, score), estimator) in enumerate(
            zip(self.scores_, self.estimators_)
        ):
            proba[:, i] = estimator.predict_proba(x) * score
            score_sum += score
        return proba.sum(axis=1) / score_sum

    def predict(self, x):
        """
        Compute the class label for the samples in x.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_timesteps) or (n_samples, n_dim, n_timesteps)
            The input samples.

        Returns
        -------
        ndarray of shape (n_samples, )
            The class label for each sample.
        """
        proba = self.predict_proba(x)
        return np.take(self.classes_, np.argmax(proba, axis=1))



import math
import numbers
import warnings

import numpy as np
from sklearn.utils._param_validation import (
    Interval,
    StrOptions,
    validate_params,
)
from sklearn.utils.validation import _is_arraylike, check_scalar

from ..utils import _safe_jagged_array
from ..utils.validation import _check_ts_array, check_array, check_option, check_type
from ._cdistance import (
    CallableMetric,
    ScaledSubsequenceMetricWrap,
    SubsequenceMetricWrap,
    _argmin_distance,
    _argmin_subsequence_distance,
    _dilated_distance_profile,
    _distance_profile,
    _paired_distance,
    _paired_subsequence_distance,
    _paired_subsequence_match,
    _pairwise_distance,
    _pairwise_subsequence_distance,
    _singleton_pairwise_distance,
    _subsequence_match,
)
from ._elastic import (
    AmercingDtwMetric,
    AmercingDtwSubsequenceMetric,
    DerivativeDtwMetric,
    DerivativeDtwSubsequenceMetric,
    DtwMetric,
    DtwSubsequenceMetric,
    EdrMetric,
    EdrSubsequenceMetric,
    ErpMetric,
    ErpSubsequenceMetric,
    LcssMetric,
    LcssSubsequenceMetric,
    MsmMetric,
    MsmSubsequenceMetric,
    ScaledDtwSubsequenceMetric,
    TweMetric,
    TweSubsequenceMetric,
    WeightedDerivativeDtwMetric,
    WeightedDerivativeDtwSubsequenceMetric,
    WeightedDtwMetric,
    WeightedDtwSubsequenceMetric,
    WeightedLcssMetric,
)
from ._mass import ScaledMassSubsequenceMetric

# from . import _cdistance, _elastic, _mass, _metric
from ._metric import (
    AngularMetric,
    AngularSubsequenceMetric,
    ChebyshevMetric,
    ChebyshevSubsequenceMetric,
    CosineMetric,
    CosineSubsequenceMetric,
    EuclideanMetric,
    EuclideanSubsequenceMetric,
    ManhattanMetric,
    ManhattanSubsequenceMetric,
    MinkowskiMetric,
    MinkowskiSubsequenceMetric,
    NormalizedEuclideanMetric,
    NormalizedEuclideanSubsequenceMetric,
    ScaledEuclideanSubsequenceMetric,
)


def _subsequence_metric_wrap(Metric):
    def f(**metric_params):
        return ScaledSubsequenceMetricWrap(Metric(**metric_params))

    return f


def _callable_metric(f):
    def wrap(**metric_params):
        return CallableMetric(f)

    return wrap


def _callable_subsequence_metric(f, scale=False):
    if scale:

        def wrap(**metric_params):
            return ScaledSubsequenceMetricWrap(CallableMetric(f))

    else:

        def wrap(**metric_params):
            return SubsequenceMetricWrap(CallableMetric(f))

    return wrap


_SUBSEQUENCE_METRICS = {
    "euclidean": EuclideanSubsequenceMetric,
    "scaled_euclidean": ScaledEuclideanSubsequenceMetric,
    "normalized_euclidean": NormalizedEuclideanSubsequenceMetric,
    "adtw": AmercingDtwSubsequenceMetric,
    "scaled_adtw": _subsequence_metric_wrap(AmercingDtwMetric),
    "wdtw": WeightedDtwSubsequenceMetric,
    "scaled_wdtw": _subsequence_metric_wrap(WeightedDtwMetric),
    "ddtw": DerivativeDtwSubsequenceMetric,
    "scaled_ddtw": _subsequence_metric_wrap(DerivativeDtwMetric),
    "wddtw": WeightedDerivativeDtwSubsequenceMetric,
    "scaled_wddtw": _subsequence_metric_wrap(WeightedDerivativeDtwMetric),
    "dtw": DtwSubsequenceMetric,
    "scaled_dtw": ScaledDtwSubsequenceMetric,
    "lcss": LcssSubsequenceMetric,
    "scaled_lcss": _subsequence_metric_wrap(LcssMetric),
    "edr": EdrSubsequenceMetric,
    "scaled_edr": _subsequence_metric_wrap(EdrMetric),
    "twe": TweSubsequenceMetric,
    "scaled_twe": _subsequence_metric_wrap(TweMetric),
    "msm": MsmSubsequenceMetric,
    "scaled_msm": _subsequence_metric_wrap(MsmMetric),
    "erp": ErpSubsequenceMetric,
    "scaled_erp": _subsequence_metric_wrap(ErpMetric),
    "mass": ScaledMassSubsequenceMetric,
    "manhattan": ManhattanSubsequenceMetric,
    "scaled_manhattan": _subsequence_metric_wrap(ManhattanMetric),
    "minkowski": MinkowskiSubsequenceMetric,
    "scaled_minkowski": _subsequence_metric_wrap(MinkowskiMetric),
    "chebyshev": ChebyshevSubsequenceMetric,
    "scaled_chebyshev": _subsequence_metric_wrap(ChebyshevMetric),
    "cosine": CosineSubsequenceMetric,
    "scaled_cosine": _subsequence_metric_wrap(CosineMetric),
    "angular": AngularSubsequenceMetric,
    "scaled_angular": _subsequence_metric_wrap(AngularMetric),
}

_METRICS = {
    "euclidean": EuclideanMetric,
    "normalized_euclidean": NormalizedEuclideanMetric,
    "adtw": AmercingDtwMetric,
    "dtw": DtwMetric,
    "ddtw": DerivativeDtwMetric,
    "wdtw": WeightedDtwMetric,
    "wddtw": WeightedDerivativeDtwMetric,
    "lcss": LcssMetric,
    "wlcss": WeightedLcssMetric,
    "erp": ErpMetric,
    "edr": EdrMetric,
    "msm": MsmMetric,
    "twe": TweMetric,
    "manhattan": ManhattanMetric,
    "minkowski": MinkowskiMetric,
    "chebyshev": ChebyshevMetric,
    "cosine": CosineMetric,
    "angular": AngularMetric,
}


def _infer_scaled_metric(metric):
    if metric.startswith("scaled_"):
        return metric
    else:
        return "scaled_" + metric


def check_metric(metric):
    if callable(metric):
        return _callable_metric(metric)
    elif metric in _METRICS:
        return _METRICS[metric]
    else:
        raise ValueError(
            "unsupported metric {}, 'metric' must be callable or a str among {}".format(
                metric, set(_METRICS.keys())
            )
        )


def check_subsequence_metric(metric, scale=False):
    if callable(metric):
        return _callable_subsequence_metric(metric, scale=scale)
    else:
        if scale:
            metric = _infer_scaled_metric(metric)

        if metric in _SUBSEQUENCE_METRICS:
            return _SUBSEQUENCE_METRICS[metric]
        else:
            raise ValueError(
                (
                    "unsupported metric '{}', 'metric' "
                    "must be callable or a str among {}"
                ).format(metric, set(_SUBSEQUENCE_METRICS.keys()))
            )


def _std_below_mean(s):
    def f(x):
        return max(np.mean(x) - s * np.std(x), np.min(x))

    return f


_THRESHOLD = {"auto": _std_below_mean(2.0)}


def _validate_subsequence(y):
    if isinstance(y, np.ndarray) and y.dtype != object:
        if y.ndim == 1:
            return [y.astype(float)]
        elif y.ndim == 2:
            y = list(y.astype(float))
        else:
            raise ValueError(
                "Expected 2D array, got {}D array instead:\narray={}.\n".format(
                    y.ndim, y
                )
            )
    elif any(_is_arraylike(e) for e in y):
        y = [np.array(e, dtype=np.double) for e in y]
    else:
        y = [np.array(y, dtype=np.double)]

    return y


def _any_in_exclude(lst, i, exclude):
    for e in lst:
        if not (e <= i - exclude or e >= i + exclude):
            return True
    return False


def _exclude_trivial_matches(indicies, distances, exclude):
    indicies_tmp = []
    distances_tmp = []
    for index, distance in zip(indicies, distances):
        if index is None:
            indicies_tmp.append(None)
            distances_tmp.append(None)
        else:
            # For each index if index has neighbors do not include those
            sort = np.argsort(distance)
            idx = np.zeros(sort.size, dtype=bool)
            excluded = []
            for i in range(index.size):
                if not _any_in_exclude(excluded, index[sort[i]], exclude):
                    excluded.append(index[sort[i]])
                    idx[sort[i]] = True

            idx = np.array(idx)
            indicies_tmp.append(index[idx])
            distances_tmp.append(distance[idx])

    return indicies_tmp, distances_tmp


def _filter_by_max_matches(indicies, distances, max_matches):
    indicies_tmp = []
    distances_tmp = []
    for index, distance in zip(indicies, distances):
        if index is None:
            indicies_tmp.append(None)
            distances_tmp.append(None)
        else:
            idx = np.argsort(distance)[:max_matches]
            indicies_tmp.append(index[idx])
            distances_tmp.append(distance[idx])

    return indicies_tmp, distances_tmp


def _filter_by_max_dist(indicies, distances, max_dist):
    indicies_tmp = []
    distances_tmp = []
    for index, distance in zip(indicies, distances):
        if index is None:
            indicies_tmp.append(None)
            distances_tmp.append(None)
        else:
            idx = max_dist(distance)
            indicies_tmp.append(index[idx])
            distances_tmp.append(distance[idx])

    return indicies_tmp, distances_tmp


def _format_return(x, y_dims, x_dims):
    if x_dims == 1 and y_dims == 1 and x.size == 1:
        return x.item()
    elif x_dims == 1 or y_dims == 1:
        return np.squeeze(x)
    else:
        return x


def pairwise_subsequence_distance(
    y,
    x,
    *,
    dim=0,
    metric="euclidean",
    metric_params=None,
    scale=False,
    return_index=False,
    n_jobs=None,
):
    """
    Minimum subsequence distance between subsequences and time series.

    Parameters
    ----------
    y : list or ndarray of shape (n_subsequences, n_timestep)
        Input time series.

        - if list, a list of array-like of shape (n_timestep, ).
    x : ndarray of shape (n_timestep, ), (n_samples, n_timestep)\
    or (n_samples, n_dims, n_timestep)
        The input data.
    dim : int, optional
        The dim to search for subsequence.
    metric : str or callable, optional
        The distance metric

        See ``_SUBSEQUENCE_METRICS.keys()`` for a list of supported metrics.
    metric_params : dict, optional
        Parameters to the metric.

        Read more about the parameters in the
        :ref:`User guide <list_of_subsequence_metrics>`.
    scale : bool, optional
        If True, scale the subsequences before distance computation.

        .. versionadded:: 1.3
    return_index : bool, optional
        - if True return the index of the best match. If there are many equally good
          matches, the first match is returned.
    n_jobs : int, optional
        The number of parallel jobs.

    Returns
    -------
    dist : float, ndarray
        The minumum distance. Return depends on input:

        - if len(y) > 1 and x.ndim > 1, return an array of shape
          (n_samples, n_subsequences).
        - if len(y) == 1, return an array of shape (n_samples, ).
        - if x.ndim == 1, return an array of shape (n_subsequences, ).
        - if x.ndim == 1 and len(y) == 1, return scalar.

    indices : int, ndarray, optional
         The start index of the minumum distance. Return dependes on input:

        - if len(y) > 1 and x.ndim > 1, return an array of shape
          (n_samples, n_subsequences).
        - if len(y) == 1, return an array of shape (n_samples, ).
        - if x.ndim == 1, return an array of shape (n_subsequences, ).
        - if x.ndim == 1 and len(y) == 1, return scalar.

    Warnings
    --------
    Passing a callable to the `metric` parameter has a significant performance
    implication.
    """
    y = _validate_subsequence(y)
    x = check_array(x, allow_3d=True, ensure_2d=False, dtype=np.double)
    for s in y:
        if s.shape[0] > x.shape[-1]:
            raise ValueError(
                "Invalid subsequnce shape (%d > %d)" % (s.shape[0], x.shape[-1])
            )

    Metric = check_subsequence_metric(metric, scale=scale)

    metric_params = metric_params or {}
    min_dist, min_ind = _pairwise_subsequence_distance(
        y,
        _check_ts_array(x),
        dim,
        Metric(**metric_params),
        n_jobs,
    )
    if return_index:
        return (
            _format_return(min_dist, len(y), x.ndim),
            _format_return(min_ind, len(y), x.ndim),
        )
    else:
        return _format_return(min_dist, len(y), x.ndim)


def paired_subsequence_distance(
    y,
    x,
    *,
    dim=0,
    metric="euclidean",
    metric_params=None,
    scale=False,
    return_index=False,
    n_jobs=None,
):
    """
    Minimum subsequence distance between the i:th subsequence and time series.

    Parameters
    ----------
    y : list or ndarray of shape (n_samples, m_timestep)
        Input time series.

        - if list, a list of array-like of shape (m_timestep, ).
    x : ndarray of shape (n_timestep, ), (n_samples, n_timestep)\
    or (n_samples, n_dims, n_timestep)
        The input data.
    dim : int, optional
        The dim to search for shapelets.
    metric : str or callable, optional
        The distance metric

        See ``_SUBSEQUENCE_METRICS.keys()`` for a list of supported metrics.
    metric_params : dict, optional
        Parameters to the metric.

        Read more about the parameters in the
        :ref:`User guide <list_of_subsequence_metrics>`.
    scale : bool, optional
        If True, scale the subsequences before distance computation.

        .. versionadded:: 1.3
    return_index : bool, optional
        - if True return the index of the best match. If there are many equally good
          matches, the first match is returned.
    n_jobs : int, optional
        The number of parallel jobs to run.

    Returns
    -------
    dist : float, ndarray
        An array of shape (n_samples, ) with the minumum distance between the i:th
        subsequence and the i:th sample.
    indices : int, ndarray, optional
        An array of shape (n_samples, ) with the index of the best matching position
        of the i:th subsequence and the i:th sample.

    Warnings
    --------
    Passing a callable to the `metric` parameter has a significant performance
    implication.
    """
    y = _validate_subsequence(y)
    x = check_array(x, allow_3d=True, ensure_2d=False, dtype=float)

    n_dims = x.shape[1] if x.ndim == 3 else 1
    if not 0 >= dim < n_dims:
        raise ValueError("The parameter dim must be 0 <= dim < n_dims")

    for s in y:
        if s.shape[0] > x.shape[-1]:
            raise ValueError(
                "Invalid subsequnce shape (%d > %d)" % (s.shape[0], x.shape[-1])
            )

    Metric = check_subsequence_metric(metric, scale=scale)

    n_samples = x.shape[0] if x.ndim > 1 else 1
    if len(y) != n_samples:
        raise ValueError(
            "The number of subsequences and samples must be the same, got %d "
            "subsequences and %d samples." % (len(y), n_samples)
        )

    metric_params = metric_params if metric_params is not None else {}
    min_dist, min_ind = _paired_subsequence_distance(
        y, _check_ts_array(x), dim, Metric(**metric_params)
    )
    if return_index:
        return (
            _format_return(min_dist, len(y), x.ndim),
            _format_return(min_ind, len(y), x.ndim),
        )
    else:
        return _format_return(min_dist, len(y), x.ndim)


def subsequence_match(  # noqa: PLR0912, PLR0915
    y,
    x,
    threshold=None,
    *,
    dim=0,
    metric="euclidean",
    metric_params=None,
    scale=False,
    max_matches=None,
    exclude=None,
    return_distance=False,
    n_jobs=None,
):
    """
    Find matching subsequnces.

    Find the positions where the distance is less than the threshold between
    the subsequence and all time series.

    - If a `threshold` is given, the default behaviour is to return all matching
      indices in the order of occurrence
    - If no `threshold` is given, the default behaviour is to return the top 10
      matching indicies ordered by distance
    - If both `threshold` and `max_matches` are given, the top matches are returned
      ordered by distance.

    Parameters
    ----------
    y : array-like of shape (yn_timestep, )
        The subsequence.
    x : ndarray of shape (n_timestep, ), (n_samples, n_timestep)\
    or (n_samples, n_dims, n_timestep)
        The input data.
    threshold : {"auto"}, float or callable, optional
        The distance threshold used to consider a subsequence matching. If no threshold
        is selected, `max_matches` defaults to 10.

        - if float, return all matches closer than threshold
        - if callable, return all matches closer than the treshold computed by the
          threshold function, given all distances to the subsequence
        - if str, return all matches according to the named threshold.
    dim : int, optional
        The dim to search for shapelets.
    metric : str or callable, optional
        The distance metric

        See ``_SUBSEQUENCE_METRICS.keys()`` for a list of supported metrics.
    metric_params : dict, optional
        Parameters to the metric.

        Read more about the parameters in the
        :ref:`User guide <list_of_subsequence_metrics>`.
    scale : bool, optional
        If True, scale the subsequences before distance computation.

        .. versionadded:: 1.3
    max_matches : int, optional
        Return the top `max_matches` matches below `threshold`.
    exclude : float or int, optional
        Exclude trivial matches in the vicinity of the match.

        - if float, the exclusion zone is computed as ``math.ceil(exclude * y.size)``
        - if int, the exclusion zone is exact

        A match is considered trivial if a match with lower distance is within `exclude`
        timesteps of another match with higher distance.
    return_distance : bool, optional
        - if True, return the distance of the match.
    n_jobs : int, optional
        The number of parallel jobs to run.

    Returns
    -------
    indicies : ndarray of shape (n_samples, ) or (n_matches, )
        The start index of matching subsequences. Returns a single array of
        n_matches if x.ndim == 1. If no matches are found for a sample, the
        array element is None.
    distance : ndarray of shape (n_samples, ), optional
        The distances of matching subsequences. Returns a single array of
        n_matches if x.ndim == 1. If no matches are found for a sample, the
        array element is None.

    Warnings
    --------
    Passing a callable to the `metric` parameter has a significant performance
    implication.
    """
    y = _validate_subsequence(y)
    if len(y) > 1:
        raise ValueError("A single subsequence expected, got %d" % len(y))

    y = y[0]
    x = check_array(x, allow_3d=True, ensure_2d=False, dtype=float)

    if y.shape[0] > x.shape[-1]:
        raise ValueError(
            "Invalid subsequnce shape (%d > %d)" % (y.shape[0], x.shape[-1])
        )

    n_dims = x.shape[1] if x.ndim == 3 else 1
    if not 0 >= dim < n_dims:
        raise ValueError("The parameter dim must be 0 <= dim < n_dims")

    Metric = check_subsequence_metric(metric, scale=scale)
    metric_params = metric_params if metric_params is not None else {}

    if n_jobs is not None:
        warnings.warn("n_jobs is not yet supported.", UserWarning)

    if threshold is None:
        threshold = np.inf
        if max_matches is None:
            max_matches = 10

    if callable(threshold):
        threshold_fn = threshold

        def max_dist(d):
            return d <= threshold_fn(d)

        threshold = np.inf
    elif isinstance(threshold, str):
        # TODO(1.3)
        if threshold == "best":
            warnings.warn(
                "threshold 'best' has been renamed to 'auto' in 1.2 "
                "and will be removed in 1.3.",
                UserWarning,
            )
            threshold = "auto"

        threshold_fn = check_option(_THRESHOLD, threshold, "threshold")

        def max_dist(d):
            return d <= threshold_fn(d)

        threshold = np.inf
    elif not isinstance(threshold, numbers.Real):
        raise TypeError(
            "threshold must be str, callable or float, not %s"
            % type(threshold).__qualname__
        )
    else:
        max_dist = None

    check_type(exclude, "exclude", (numbers.Integral, numbers.Real), required=False)
    if isinstance(exclude, numbers.Integral):
        check_scalar(exclude, "exclude", numbers.Integral, min_val=0)
    elif isinstance(exclude, numbers.Real):
        check_scalar(
            exclude,
            "exclude",
            numbers.Real,
            min_val=0,
        )
        exclude = math.ceil(y.size * exclude)

    indices, distances = _subsequence_match(
        y,
        _check_ts_array(x),
        threshold,
        dim,
        Metric(**metric_params),
        n_jobs,
    )

    if max_dist is not None:
        indices, distances = _filter_by_max_dist(indices, distances, max_dist)

    if exclude:
        indices, distances = _exclude_trivial_matches(indices, distances, exclude)

    if max_matches:
        indices, distances = _filter_by_max_matches(indices, distances, max_matches)

    indices = _format_return(_safe_jagged_array(indices), len(y), x.ndim)
    if indices.size == 1:
        indices = indices.item()

    if return_distance:
        distances = _format_return(_safe_jagged_array(distances), len(y), x.ndim)
        if distances.size == 1:
            distances = distances.item()

        return indices, distances
    else:
        return indices


def paired_subsequence_match(  # noqa: PLR0912
    y,
    x,
    threshold=None,
    *,
    dim=0,
    metric="euclidean",
    metric_params=None,
    scale=False,
    max_matches=None,
    return_distance=False,
    n_jobs=None,
):
    """
    Find matching subsequnces.

    Find the positions where the distance is less than the threshold between
    the i:th subsequences and time series.

    - If a `threshold` is given, the default behaviour is to return all matching
      indices in the order of occurrence
    - If no `threshold` is given, the default behaviour is to return the top 10
      matching indicies ordered by distance
    - If both `threshold` and `max_matches` are given, the top matches are returned
      ordered by distance and time series.

    Parameters
    ----------
    y : list or ndarray of shape (n_samples, n_timestep)
        Input time series.

        - if list, a list of array-like of shape (n_timestep, ) with length n_samples.
    x : ndarray of shape (n_samples, n_timestep) or (n_samples, n_dims, n_timestep)
        The input data.
    threshold : float, optional
        The distance threshold used to consider a subsequence matching. If no threshold
        is selected, `max_matches` defaults to 10.
    dim : int, optional
        The dim to search for shapelets.
    metric : str or callable, optional
        The distance metric

        See ``_SUBSEQUENCE_METRICS.keys()`` for a list of supported metrics.
    metric_params : dict, optional
        Parameters to the metric.

        Read more about the parameters in the
        :ref:`User guide <list_of_subsequence_metrics>`.
    scale : bool, optional
        If True, scale the subsequences before distance computation.

        .. versionadded:: 1.3
    max_matches : int, optional
        Return the top `max_matches` matches below `threshold`.

        - If a `threshold` is given, the default behaviour is to return all matching
          indices in the order of occurrence .
        - If no `threshold` is given, the default behaviour is to return the top 10
          matching indicies ordered by distance
        - If both `threshold` and `max_matches` are given the top matches are returned
          ordered by distance.
    return_distance : bool, optional
        If True, return the distance of the match.
    n_jobs : int, optional
        The number of parallel jobs to run. Ignored.

    Returns
    -------
    indicies : ndarray of shape (n_samples, )
        The start index of matching subsequences.
    distance : ndarray of shape (n_samples, ), optional
        The distances of matching subsequences.

    Warnings
    --------
    Passing a callable to the `metric` parameter has a significant performance
    implication.
    """
    y = _validate_subsequence(y)
    x = check_array(x, allow_3d=True, dtype=np.double)
    if len(y) != x.shape[0]:
        raise ValueError("x and y must have the same number of samples")

    n_dims = x.shape[1] if x.ndim == 3 else 1
    if not 0 >= dim < n_dims:
        raise ValueError("The parameter dim must be 0 <= dim < n_dims")

    for s in y:
        if s.shape[0] > x.shape[-1]:
            raise ValueError(
                "invalid subsequnce shape (%d > %d)" % (s.shape[0], x.shape[-1])
            )

    Metric = check_subsequence_metric(metric, scale=scale)
    metric_params = metric_params if metric_params is not None else {}

    if n_jobs is not None:
        warnings.warn("n_jobs is not yet supported.", UserWarning)

    if threshold is None:
        threshold = np.inf
        if max_matches is None:
            max_matches = 10

    if callable(threshold):
        threshold_fn = threshold

        def max_dist(d):
            return d <= threshold_fn(d)

        threshold = np.inf
    elif isinstance(threshold, str):
        threshold_fn = check_option(_THRESHOLD, threshold, "threshold")

        def max_dist(d):
            return d <= threshold_fn(d)

        threshold = np.inf
    elif not isinstance(threshold, numbers.Real):
        raise TypeError(
            "threshold must be str, callable or float, not %s"
            % type(threshold).__qualname__
        )
    else:
        max_dist = None

    indices, distances = _paired_subsequence_match(
        y,
        _check_ts_array(x),
        threshold,
        dim,
        Metric(**metric_params),
        n_jobs,
    )

    if max_dist is not None:
        indices, distances = _filter_by_max_dist(indices, distances, max_dist)

    if max_matches:
        indices, distances = _filter_by_max_matches(indices, distances, max_matches)

    indices = _format_return(_safe_jagged_array(indices), len(y), x.ndim)
    if indices.size == 1:
        indices = indices.reshape(1)

    if return_distance:
        distances = _format_return(_safe_jagged_array(distances), len(y), x.ndim)
        if distances.size == 1:
            distances = distances.reshape(1)

        return indices, distances
    else:
        return indices


def paired_distance(  # noqa: PLR0912
    x,
    y,
    *,
    dim="warn",
    metric="euclidean",
    metric_params=None,
    n_jobs=None,
):
    """
    Compute the distance between the i:th time series.

    Parameters
    ----------
    x : ndarray of shape (n_samples, n_timestep) or (n_samples, n_dims, n_timestep)
        The input data.
    y : ndarray of shape (n_samples, n_timestep) or (n_samples, n_dims, n_timestep)
        The input data. y will be broadcasted to the shape of x.
    dim : int or {'mean', 'full'}, optional
        The dim to compute distance.
    metric : str or callable, optional
        The distance metric

        See ``_METRICS.keys()`` for a list of supported metrics.
    metric_params : dict, optional
        Parameters to the metric.

        Read more about the parameters in the
        :ref:`User guide <list_of_metrics>`.
    n_jobs : int, optional
        The number of parallel jobs.

    Returns
    -------
    ndarray
        The distances. Return depends on input:

        - if x.ndim == 1, return scalar.
        - if dim='full', return ndarray of shape (n_dims, n_samples).
        - if x.ndim > 1, return an ndarray of shape (n_samples, ).

    Warnings
    --------
    Passing a callable to the `metric` parameter has a significant performance
    implication.
    """
    x = check_array(x, allow_3d=True, ensure_2d=False, dtype=float)
    y = check_array(y, allow_3d=True, ensure_2d=False, dtype=float)
    x, y = np.broadcast_arrays(x, y)
    if x.ndim != y.ndim:
        raise ValueError(
            "x (%dD-array) and y (%dD-array) are not compatible." % (x.ndim, y.ndim)
        )

    if x.ndim == 3 and x.shape[1] != y.shape[1]:
        raise ValueError("x and y must have the same number of dimensions.")

    if x.ndim > 1 and y.ndim > 1 and x.shape[0] != y.shape[0]:
        raise ValueError("x and y must have the same number of samples.")

    if n_jobs is not None:
        warnings.warn("n_jobs is not yet supported.", UserWarning)

    Metric = check_metric(metric)

    metric_params = metric_params if metric_params is not None else {}
    metric = Metric(**metric_params)
    if x.shape[x.ndim - 1] != x.shape[x.ndim - 1] and not metric.is_elastic:
        raise ValueError(
            "Illegal n_timestep (%r != %r) for non-elastic distance measure"
            % (x.shape[x.ndim - 1], y.shape[y.ndim - 1])
        )

    n_dims = x.shape[1] if x.ndim == 3 else 1

    # TODO(1.3)
    if dim == "warn":
        if n_dims > 1:
            warnings.warn(
                "The default value for dim will change to 'mean' from 0 in 1.3. "
                "Explicitly set dim=0 to keep the current behaviour for 3d-arrays.",
                DeprecationWarning,
            )
        dim = 0

    if n_dims == 1 and dim == "mean":
        dim = 0

    x_ = _check_ts_array(x)
    y_ = _check_ts_array(y)
    if dim in ["mean", "full"]:
        distances = [_paired_distance(x_, y_, d, metric, n_jobs) for d in range(n_dims)]

        if dim == "mean":
            distances = np.mean(distances, axis=0)
        else:
            distances = np.stack(distances, axis=0)

    elif isinstance(dim, numbers.Integral) and 0 <= dim < n_dims:
        distances = _paired_distance(x_, y_, dim, metric, n_jobs)
    else:
        raise ValueError("The parameter dim must be 0 <= dim < n_dims")

    return _format_return(distances, y.ndim, x.ndim)


def pairwise_distance(  # noqa: PLR0912, PLR0915
    x,
    y=None,
    *,
    dim="warn",
    metric="euclidean",
    metric_params=None,
    n_jobs=None,
):
    """
    Compute the distance between subsequences and time series.

    Parameters
    ----------
    x : ndarray of shape (n_timestep, ), (x_samples, n_timestep) or \
            (x_samples, n_dims, n_timestep)
        The input data.
    y : ndarray of shape (n_timestep, ), (y_samples, n_timestep) or \
            (y_samples, n_dims, n_timestep), optional
        The input data.
    dim : int or {'mean', 'full'}, optional
        The dim to compute distance.
    metric : str or callable, optional
        The distance metric

        See ``_METRICS.keys()`` for a list of supported metrics.
    metric_params : dict, optional
        Parameters to the metric.

        Read more about the parameters in the
        :ref:`User guide <list_of_metrics>`.
    n_jobs : int, optional
        The number of parallel jobs.

    Returns
    -------
    float or ndarray
        The distances. Return depends on input.

        - if x.ndim == 1 and y.ndim == 1, scalar.
        - if dim="full", array of shape (n_dims, x_samples, y_samples).
        - if dim="full" and y is None, array of shape (n_dims, x_samples, x_samples).
        - if x.ndim > 1 and y is None, array of shape (x_samples, x_samples).
        - if x.ndim > 1 and y.ndim > 1, array of shape (x_samples, y_samples).
        - if x.ndim == 1 and y.ndim > 1, array of shape (y_samples, ).
        - if y.ndim == 1 and x.ndim > 1, array of shape (x_samples, ).

    Warnings
    --------
    Passing a callable to the `metric` parameter has a significant performance
    implication.
    """
    Metric = check_metric(metric)
    metric_params = metric_params if metric_params is not None else {}
    metric = Metric(**metric_params)

    if y is None:
        y = x

    if x is y:
        x = check_array(x, allow_3d=True, ensure_2d=False, dtype=float)
        if x.ndim == 1:
            return 0.0

        x_ = _check_ts_array(x)
        n_dims = x.shape[1] if x.ndim == 3 else 1

        # TODO(1.3)
        if dim == "warn":
            if n_dims > 1:
                warnings.warn(
                    "The default value for dim will change to 'mean' from 0 in 1.3. "
                    "Explicitly set dim=0 to keep the current behaviour for 3d-arrays.",
                    DeprecationWarning,
                )
            dim = 0

        if n_dims == 1 and dim == "mean":
            dim = 0

        if dim in ["mean", "full"]:
            distances = [
                _singleton_pairwise_distance(x_, d, metric, n_jobs)
                for d in range(n_dims)
            ]

            if dim == "mean":
                distances = np.mean(distances, axis=0)
            else:
                distances = np.stack(distances, axis=0)

        elif isinstance(dim, numbers.Integral) and 0 <= dim < n_dims:
            distances = _singleton_pairwise_distance(x_, dim, metric, n_jobs)
        else:
            raise ValueError("The parameter dim must be 0 <= dim < n_dims")

        return distances
    else:
        x = check_array(x, allow_3d=True, ensure_2d=False, dtype=np.double)
        y = check_array(y, allow_3d=True, ensure_2d=False, dtype=np.double)
        if x.ndim != 1 and y.ndim != 1 and x.ndim != y.ndim:
            raise ValueError(
                "x (%dD-array) and y (%dD-array) are not compatible" % (x.ndim, y.ndim)
            )

        if x.ndim == 3 and x.shape[1] != y.shape[1]:
            raise ValueError("x and y must have the same number of dimensions.")

        if x.shape[-1] != y.shape[-1] and not metric.is_elastic:
            raise ValueError(
                "Illegal n_timestep (%r != %r) for non-elastic distance measure"
                % (x.shape[-1], y.shape[-1])
            )

        x_ = _check_ts_array(x)
        y_ = _check_ts_array(y)
        n_dims = x.shape[1] if x.ndim == 3 else 1

        # TODO(1.3)
        if dim == "warn":
            if n_dims > 1:
                warnings.warn(
                    "The default value for dim will change to 'mean' from 0 in 1.3. "
                    "Explicitly set dim=0 to keep the current behaviour for 3d-arrays.",
                    DeprecationWarning,
                )
            dim = 0

        if n_dims == 1 and dim == "mean":
            dim = 0

        if dim in ["mean", "full"]:
            distances = [
                _pairwise_distance(x_, y_, d, metric, n_jobs) for d in range(n_dims)
            ]

            if dim == "mean":
                distances = np.mean(distances, axis=0)
            else:
                distances = np.stack(distances, axis=0)

        elif isinstance(dim, numbers.Integral) and 0 <= dim < n_dims:
            distances = _pairwise_distance(x_, y_, dim, metric, n_jobs)
        else:
            raise ValueError("The parameter dim must be 0 <= dim < n_dims")

        return _format_return(distances, y.ndim, x.ndim)


@validate_params(
    {
        "x": ["array-like"],
        "y": [None, "array-like"],
        "k": [Interval(numbers.Integral, 1, None, closed="left")],
        "metric": [callable, StrOptions(_METRICS.keys())],
        "metric_params": [None, dict],
        "return_distance": [bool],
        "sorted": [bool],
        "n_jobs": [numbers.Integral, None],
    },
    prefer_skip_nested_validation=True,
)
def argmin_distance(
    x,
    y=None,
    *,
    dim=0,
    k=1,
    metric="euclidean",
    metric_params=None,
    sorted=False,
    return_distance=False,
    n_jobs=None,
):
    """
    Find the indicies of the samples with the lowest distance in `Y`.

    Parameters
    ----------
    x : univariate time-series or multivariate time-series
        The needle.
    y : univariate time-series or multivariate time-series, optional
        The haystack.
    dim : int, optional
        The dimension where the distance is computed.
    k : int, optional
        The number of closest samples.
    metric : str, optional
        The distance metric

        See ``_METRICS.keys()`` for a list of supported metrics.
    metric_params : dict, optional
        Parameters to the metric.

        Read more about the parameters in the
        :ref:`User guide <list_of_metrics>`.
    sorted : bool, optional
        Sort the indicies from smallest to largest distance.
    return_distance : bool, optional
        Return the distance for the `k` samples.
    n_jobs : int, optional
        The number of parallel jobs.

    Returns
    -------
    indices : ndarray of shape (n_samples, k)
        The indices of the samples in `Y` with the smallest distance.
    distance : ndarray of shape (n_samples, k), optional
        The distance of the samples in `Y` with the smallest distance.

    Warnings
    --------
    Passing a callable to the `metric` parameter has a significant performance
    implication.

    Examples
    --------
    >>> from wildoar.distance import argmin_distance
    >>> X = np.array([[1, 2, 3, 4], [10, 1, 2, 3]])
    >>> Y = np.array([[1, 2, 11, 2], [2, 4, 6, 7], [10, 11, 2, 3]])
    >>> argmin_distance(X, Y, k=2, return_distance=True)
    (array([[0, 1],
            [1, 2]]),
     array([[ 8.24621125,  4.79583152],
            [10.24695077, 10.        ]]))
    """
    metric_params = metric_params if metric_params is not None else {}
    metric = check_metric(metric)(**metric_params)

    x = check_array(
        x, allow_3d=True, ensure_2d=False, ensure_ts_array=True, dtype=float
    )
    if y is None:
        y = x
    else:
        y = check_array(
            y, allow_3d=True, ensure_2d=False, ensure_ts_array=True, dtype=float
        )

    if x.ndim not in (1, y.ndim):
        raise ValueError(
            f"x ({x.ndim}d-array) and y ({y.ndim}d-array) are not compatible."
        )

    if x.shape[-1] != y.shape[-1] and not metric.is_elastic:
        raise ValueError(
            "Illegal n_timestep (%d != %d) for non-elastic distance measure."
            % (x.shape[-1], y.shape[-1])
        )

    n_dims = x.shape[1] if x.ndim == 3 else 1
    k = min(k, y.shape[0])
    if 0 <= dim < 1:
        indices, distances = _argmin_distance(x, y, dim, metric, k, n_jobs)

        if sorted:
            sort = np.argsort(distances, axis=1)
            indices = np.take_along_axis(indices, sort, axis=1)
            if return_distance:
                distances = np.take_along_axis(distances, sort, axis=1)

        if return_distance:
            return indices, distances
        else:
            return indices
    else:
        raise ValueError(f"The parameter dim must be dim ({dim}) < n_dims ({n_dims})")


@validate_params(
    {
        "y": ["array-like"],
        "x": ["array-like"],
        "dilation": [Interval(numbers.Integral, 1, None, closed="left")],
        "padding": [
            Interval(numbers.Integral, 0, None, closed="left"),
            StrOptions({"same"}),
        ],
        "dim": [Interval(numbers.Integral, 0, None, closed="left")],
        "metric": [callable, StrOptions(_SUBSEQUENCE_METRICS.keys())],
        "metric_params": [None, dict],
        "scale": [None, bool],
        "n_jobs": [numbers.Integral, None],
    },
    prefer_skip_nested_validation=True,
)
def distance_profile(  # noqa: PLR0912
    y,
    x,
    *,
    dilation=1,
    padding=0,
    dim=0,
    metric="mass",
    metric_params=None,
    scale=False,
    n_jobs=None,
):
    """
    Compute the distance profile.

    The distance profile corresponds to the distance of the subsequences in y
    for every time point of the samples in x.

    Parameters
    ----------
    y : array-like of shape (m_timestep, ) or (n_samples, m_timestep)
        The subsequences. if `y.ndim` is 1, we will broacast `y` to have the
        same number of samples as `x`.
    x : ndarray of shape (n_timestep, ), (n_samples, n_timestep)\
    or (n_samples, n_dims, n_timestep)
        The samples. If `x.ndim` is 1, we will broadcast `x` to have the same
        number of samples as `y`.
    dilation : int, optional
        The dilation, i.e., the spacing between points in the subsequences.
    padding : int or {"same"}, optional
        The amount of padding applied to the input time series. If "same", the output
        size is the same as the input size.
    dim : int, optional
        The dim to search for shapelets.
    metric : str or callable, optional
        The distance metric

        See ``_SUBSEQUENCE_METRICS.keys()`` for a list of supported metrics.
    metric_params : dict, optional
        Parameters to the metric.

        Read more about the parameters in the
        :ref:`User guide <list_of_subsequence_metrics>`.
    scale : bool, optional
        If True, scale the subsequences before distance computation.
    n_jobs : int, optional
        The number of parallel jobs to run.

    Returns
    -------
    ndarray of shape (n_samples, output_size) or (output_size, )
        The distance profile. `output_size` is given by:
        `n_timestep + 2 * padding - (n_timestep - 1) * dilation + 1) + 1`.
        If both `x` and `y` contains a single subsequence and a single sample,
        the output is squeezed.

    Warnings
    --------
    Passing a callable to the `metric` parameter has a significant performance
    implication.

    Examples
    --------
    >>> from wildboar.datasets import load_dataset
    >>> from wildboar.distance import distance_profile
    >>> X, _ = load_dataset("ECG200")
    >>> distance_profile(X[0], X[1:].reshape(-1))
    array([14.00120332, 14.41943788, 14.81597243, ...,  4.75219094,
           5.72681005,  6.70155561])

    >>> distance_profile(
    ...     X[0, 0:9], X[1:5], metric="dtw", dilation=2, padding="same"
    ... )[0, :10]
    array([8.01881424, 7.15083281, 7.48856368, 6.83139294, 6.75595579,
           6.30073636, 6.65346307, 6.27919601, 6.25666948, 6.0961576 ])
    """
    y = check_array(y, dtype=float, ensure_2d=False).squeeze()
    x = check_array(x, allow_3d=True, ensure_2d=False, dtype=float).squeeze()

    if x.ndim == 1 and y.ndim != 1:
        x = np.broadcast_to(x, shape=(y.shape[0], x.shape[0]))

    if y.ndim == 1 and x.ndim != 1:
        y = np.broadcast_to(y, shape=(x.shape[0], y.shape[0]))

    x = _check_ts_array(x)
    y = _check_ts_array(y)

    shapelet_size = (y.shape[2] - 1) * dilation + 1
    if padding == "same":
        if y.shape[2] % 2 == 0:
            raise ValueError(
                "padding='same' is only supported for odd subsequence length"
            )
        padding = shapelet_size // 2

    input_size = x.shape[2] + 2 * padding

    if shapelet_size > input_size:
        raise ValueError("subsequence in y is larger than input in x.")

    if y.shape[0] != x.shape[0]:
        raise ValueError("y and x must have the same number of samples.")

    if dim >= x.shape[1]:
        raise ValueError(
            f"The parameter dim must be dim ({dim}) < n_dims ({x.shape[1]})"
        )

    if dilation == 1 and padding == 0:
        Metric = check_subsequence_metric(metric, scale=scale)
        metric_params = metric_params if metric_params is not None else {}
        dp = _distance_profile(y, x, dim, Metric(**metric_params), n_jobs)
    else:
        # HACK
        if metric == "mass":
            metric = "scaled_euclidean"

        scale = (isinstance(metric, str) and metric.startswith("scaled_")) or scale
        if isinstance(metric, str) and metric.startswith("scaled_"):
            metric = metric[7:]

        Metric = check_metric(metric)
        metric_params = metric_params if metric_params is not None else {}

        if scale:
            std = np.std(y, axis=-1, keepdims=True)
            mean = np.mean(y, axis=-1, keepdims=True)
            std[std < 1e-13] = 1  # NOTE! see EPSILON in _cdistance.pxd
            y = (y - mean) / std

        dp = _dilated_distance_profile(
            y,
            x,
            dim,
            Metric(**metric_params),
            dilation,
            padding,
            scale,
            n_jobs,
        )

    return np.squeeze(dp)


@validate_params(
    {
        "y": ["array-like"],
        "x": ["array-like"],
        "dim": [Interval(numbers.Integral, 0, None, closed="left")],
        "k": [Interval(numbers.Integral, 1, None, closed="left")],
        "metric": [callable, StrOptions(_SUBSEQUENCE_METRICS.keys() - {"mass"})],
        "metric_params": [None, dict],
        "scale": [None, bool],
        "return_distance": [bool],
        "n_jobs": [numbers.Integral, None],
    },
    prefer_skip_nested_validation=True,
)
def argmin_subsequence_distance(  # noqa: PLR0912
    y,
    x,
    *,
    dim=0,
    k=1,
    metric="euclidean",
    metric_params=None,
    scale=False,
    return_distance=False,
    n_jobs=None,
):
    """
    Compute the k:th closest subsequences.

    For the i:th shapelet and the i:th sample return the index and, optionally,
    the distance of the `k` closest matches.

    Parameters
    ----------
    y : array-like of shape (n_samples, m_timestep) or list of 1d-arrays
        The subsequences.
    x : array-like of shape (n_timestep, ), (n_samples, n_timestep)\
    or (n_samples, n_dims, n_timestep)
        The samples. If x.ndim == 1, it will be broadcast have the same number
        of samples that y.
    dim : int, optional
        The dimension in x to find subsequences in.
    k : int, optional
        The of closest subsequences to find.
    metric : str, optional
        The metric.

        See ``_SUBSEQUENCE_METRICS.keys()`` for a list of supported metrics.
    metric_params : dict, optional
        Parameters to the metric.

        Read more about the parameters in the
        :ref:`User guide <list_of_subsequence_metrics>`.
    scale : bool, optional
        If True, scale the subsequences before distance computation.
    return_distance : bool, optional
        Return the distance for the `k` closest subsequences.
    n_jobs : int, optional
       The number of parallel jobs.

    Returns
    -------
    indices : ndarray of shape (n_samples, k)
        The indices of the `k` closest subsequences.
    distance : ndarray of shape (n_samples, k), optional
        The distance of the `k` closest subsequences.

    Warnings
    --------
    Passing a callable to the `metric` parameter has a significant performance
    implication.

    Examples
    --------
    >>> import numpy as np
    >>> from wildboar.datasets import load_dataset
    >>> from wildboar.distance import argmin_subsequence_distance
    >>> s = np.lib.stride_tricks.sliding_window_view(X[0], window_shape=10)
    >>> x = np.broadcast_to(X[0], shape=(s.shape[0], X.shape[1]))
    >>> argmin_subsequence_distance(s, x, k=4)
    """
    if isinstance(y, np.ndarray) and y.dtype == float:
        y = check_array(y, allow_3d=False, ensure_ts_array=True)
        y_len = np.broadcast_to(y.shape[2], y.shape[0]).astype(np.intp)
    else:
        # we assume y is a list of ndarrays
        y_len = np.empty(len(y), dtype=np.intp)
        for i, shapelet in enumerate(y):
            if shapelet.ndim > 1:
                raise ValueError("shapelet must be 1d-array")

            y_len[i] = shapelet.shape[0]

        y_tmp = np.empty((len(y), 1, np.max(y_len)), dtype=float)
        for i, shapelet in enumerate(y):
            y_tmp[i, 0, : shapelet.shape[0]] = shapelet

        y = y_tmp

    if x.ndim == 1:
        x = np.broadcast_to(x, shape=(y.shape[0], x.shape[0]))

    x = check_array(x, allow_3d=True, ensure_ts_array=True)

    if not y.shape[2] <= x.shape[2]:
        raise ValueError("the longest subsequence must be shorter than samples.")

    if y.shape[0] != x.shape[0]:
        raise ValueError("both arrays must have the same number of samples.")

    if dim >= x.shape[1]:
        raise ValueError(
            f"The parameter dim must be dim ({dim}) < n_dims ({x.shape[1]})"
        )

    if not k <= (x.shape[2] - y.shape[2] + 1):
        raise ValueError("k must be less x.shape[-1] - y.shape[-1] + 1.")

    scale = (isinstance(metric, str) and metric.startswith("scaled_")) or scale
    if isinstance(metric, str) and metric.startswith("scaled_"):
        metric = metric[7:]

    Metric = check_metric(metric)
    metric_params = metric_params if metric_params is not None else {}

    indices, distances = _argmin_subsequence_distance(
        y,
        y_len,
        x,
        dim,
        Metric(**metric_params),
        k,
        scaled=bool(scale),
        n_jobs=n_jobs,
    )

    if return_distance:
        return indices, distances
    else:
        return indices



# cython: language_level=3
# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: initializedcheck=False

# Authors: Isak Samsten
# License: BSD 3 clause

import numpy as np

from libc.math cimport NAN
from libc.stdlib cimport free, malloc
from numpy cimport uint32_t

from ..utils cimport TSArray, _stats
from ..utils._misc cimport to_ndarray_double
from ..utils._rand cimport RAND_R_MAX, rand_int, shuffle
from ._attr_gen cimport Attribute, AttributeGenerator
from .catch22 cimport _catch22


cdef struct Interval:
    Py_ssize_t start
    Py_ssize_t length
    Py_ssize_t random_output


cdef class Summarizer:
    cdef void summarize_all(
        self, const double *x, Py_ssize_t length, double *out
    ) noexcept nogil:
        cdef Py_ssize_t i
        for i in range(self.n_outputs()):
            out[i] = self.summarize(i, x, length)

    cdef double summarize(
        self, Py_ssize_t i, const double *x, Py_ssize_t length
    ) noexcept nogil:
        pass

    cdef void reset(self, TSArray X) noexcept nogil:
        pass

    cdef Py_ssize_t n_outputs(self) noexcept nogil:
        return -1


cdef class MeanSummarizer(Summarizer):
    cdef double summarize(
        self, Py_ssize_t i, const double *x, Py_ssize_t length
    ) noexcept nogil:
        return _stats.mean(x, length)

    cdef Py_ssize_t n_outputs(self) noexcept nogil:
        return 1


cdef class VarianceSummarizer(Summarizer):
    cdef double summarize(
        self, Py_ssize_t i, const double *x, Py_ssize_t length
    ) noexcept nogil:
        return _stats.variance(x, length)

    cdef Py_ssize_t n_outputs(self) noexcept nogil:
        return 1


cdef class SlopeSummarizer(Summarizer):
    cdef double summarize(
        self, Py_ssize_t i, const double *x, Py_ssize_t length
    ) noexcept nogil:
        return _stats.slope(x, length)

    cdef Py_ssize_t n_outputs(self) noexcept nogil:
        return 1


cdef class MeanVarianceSlopeSummarizer(Summarizer):
    cdef double summarize(
        self, Py_ssize_t i, const double *x, Py_ssize_t length
    ) noexcept nogil:
        if i == 0:
            return _stats.mean(x, length)
        elif i == 1:
            return _stats.variance(x, length)
        elif i == 2:
            return _stats.slope(x, length)
        else:
            return NAN

    cdef Py_ssize_t n_outputs(self) noexcept nogil:
        return 3


cdef class Catch22Summarizer(Summarizer):
    cdef int *bin_count
    cdef double *bin_edges
    cdef double *ac
    cdef double *welch_f
    cdef double *welch_s
    cdef double *window

    def __cinit__(self):
        self.bin_count = <int*>malloc(sizeof(int) * 10)
        self.bin_edges = <double*>malloc(sizeof(double) * 11)
        self.ac = NULL
        self.welch_f = NULL
        self.welch_s = NULL
        self.window = NULL
        if self.bin_count == NULL or self.bin_edges == NULL:
            raise MemoryError()

    cdef void reset(self, TSArray X) noexcept nogil:
        if self.ac != NULL:
            free(self.ac)

        if self.welch_s != NULL:
            free(self.welch_s)

        if self.welch_f != NULL:
            free(self.welch_f)

        if self.window != NULL:
            free(self.window)

        cdef Py_ssize_t n_timestep = X.shape[2]
        cdef Py_ssize_t welch_length = _stats.next_power_of_2(n_timestep)
        self.ac = <double*> malloc(sizeof(double) * n_timestep)
        self.window = <double*> malloc(sizeof(double)* n_timestep)
        self.welch_f = <double*> malloc(sizeof(double) * welch_length)
        self.welch_s = <double*> malloc(sizeof(double) * welch_length)

        if (
            self.ac == NULL or
            self.welch_s == NULL or
            self.welch_f == NULL or
            self.window == NULL
        ):
            raise MemoryError()

        cdef Py_ssize_t i
        for i in range(n_timestep):
            self.window[i] = 1.0

    def __dealloc__(self):
        free(self.bin_count)
        free(self.bin_edges)
        if self.ac != NULL:
            free(self.ac)

        if self.welch_f != NULL:
            free(self.welch_f)

        if self.welch_s != NULL:
            free(self.welch_s)

        if self.window != NULL:
            free(self.window)

    def __reduce__(self):
        return self.__class__, ()

    # This is ugly
    cdef double summarize(
        self, Py_ssize_t i, const double *x, Py_ssize_t length
    ) noexcept nogil:
        cdef Py_ssize_t welch_length = -1
        cdef int n_welch = -1

        if i == 15 or i == 20:
            n_welch = _stats.welch(
                x,
                length,
                _stats.next_power_of_2(length),
                1.0,
                self.window,
                length,
                self.welch_s,
                self.welch_f
            )
        if i == 2 or i == 3 or i == 8 or i == 10 or i == 12:
            _stats.auto_correlation(x, length, self.ac)

        if i == 0:
            return _catch22.histogram_mode5(x, length, self.bin_count, self.bin_edges)
        elif i == 1:
            return _catch22.histogram_mode10(x, length, self.bin_count, self.bin_edges)
        elif i == 2:
            return _catch22.f1ecac(self.ac, length)
        elif i == 3:
            return _catch22.first_min(self.ac, length)
        elif i == 4:
            return _catch22.histogram_ami_even_2_5(x, length)
        elif i == 5:
            return _catch22.trev_1_num(x, length)
        elif i == 6:
            return _catch22.hrv_classic_pnn(x, length, 40)
        elif i == 7:
            return _catch22.above_mean_stretch(x, length)
        elif i == 8:
            return _catch22.transition_matrix_3ac_sumdiagcov(x, self.ac, length)
        elif i == 9:
            return _catch22.periodicity_wang_th0_01(x, length)
        elif i == 10:
            return _catch22.embed2_dist_tau_d_expfit_meandiff(x, self.ac, length)
        elif i == 11:
            return _catch22.auto_mutual_info_stats_gaussian_fmmi(x, length, 40)
        elif i == 12:
            return _catch22.local_mean_tauresrat(x, self.ac, length, 1)
        elif i == 13:
            return _catch22.outlier_include_np_mdrmd(x, length, 1, 0.01)
        elif i == 14:
            return _catch22.outlier_include_np_mdrmd(x, length, -1, 0.01)
        elif i == 15:
            return _catch22.summaries_welch_rect(
                x, length, 1, self.welch_s, self.welch_f, n_welch
            )
        elif i == 16:
            return _catch22.below_diff_stretch(x, length)
        elif i == 17:
            return _catch22.motif_three_quantile_hh(x, length)
        elif i == 18:
            return _catch22.fluct_anal_2_50_1_logi_prop_r1(x, length, 1, 0)
        elif i == 19:
            return _catch22.fluct_anal_2_50_1_logi_prop_r1(x, length, 2, 1)
        elif i == 20:
            return _catch22.summaries_welch_rect(
                x, length, 0, self.welch_s, self.welch_f, n_welch
            )
        elif i == 21:
            return _catch22.local_mean_std(x, length, 3)
        else:
            return NAN

    cdef Py_ssize_t n_outputs(self) noexcept nogil:
        return 22

cdef class PyFuncSummarizer(Summarizer):
    cdef list func
    cdef object x_buffer

    def __cinit__(self, func):
        self.func = func
        self.x_buffer = None

    def __reduce__(self):
        return self.__class__, (self.func, )

    cdef void reset(self, TSArray X) noexcept nogil:
        with gil:
            self.x_buffer = np.empty(X.shape[2], dtype=float)

    cdef void summarize_all(
            self,
            double *x,
            Py_ssize_t length,
            double *out
    ) noexcept nogil:
        cdef double value
        cdef Py_ssize_t i
        with gil:
            for i in range(length):
                self.x_buffer[i] = x[i]

        for i in range(self.n_outputs()):
            with gil:
                value = self.func[i](self.x_buffer[0:length])
            out[i] = value

    cdef double summarize(
        self, Py_ssize_t m, const double *x, Py_ssize_t length
    ) noexcept nogil:
        cdef double value
        cdef Py_ssize_t i
        with gil:
            for i in range(length):
                self.x_buffer[i] = x[i]
            value = self.func[m](self.x_buffer[0:length])
        return value

    cdef Py_ssize_t n_outputs(self) noexcept nogil:
        with gil:
            return len(self.func)


cdef inline Py_ssize_t _imin(Py_ssize_t a, Py_ssize_t b) noexcept nogil:
    return a if a < b else b


cdef class IntervalAttributeGenerator(AttributeGenerator):
    cdef Py_ssize_t n_intervals
    cdef Summarizer summarizer

    def __cinit__(self, Py_ssize_t n_intervals, Summarizer summarizer, *args, **kwargs):
        self.n_intervals = n_intervals
        self.summarizer = summarizer

    def __reduce__(self):
        return self.__class__, (self.n_intervals, self.summarizer)

    cdef int reset(self, TSArray X) noexcept nogil:
        self.summarizer.reset(X)
        return 0

    cdef Py_ssize_t get_n_attributess(self, TSArray X) noexcept nogil:
        return X.shape[1] * self.n_intervals

    cdef Py_ssize_t get_n_outputs(self, TSArray X) noexcept nogil:
        return self.get_n_attributess(X) * self.summarizer.n_outputs()

    cdef Py_ssize_t next_attribute(
            self,
            Py_ssize_t attribute_id,
            TSArray X,
            Py_ssize_t *samples,
            Py_ssize_t n_samples,
            Attribute *transient,
            uint32_t *seed,
    ) noexcept nogil:
        cdef Interval *interval = <Interval*> malloc(sizeof(Interval))
        interval.length = X.shape[2] // self.n_intervals
        interval.start = (attribute_id % self.n_intervals) * interval.length + _imin(
            attribute_id % self.n_intervals, X.shape[2] % self.n_intervals
        )

        if attribute_id % self.n_intervals < X.shape[2] % self.n_intervals:
            interval.length += 1

        interval.random_output = 0
        if self.summarizer.n_outputs() > 1:
            interval.random_output = rand_int(0, self.summarizer.n_outputs(), seed)

        transient.dim = attribute_id // self.n_intervals
        transient.attribute = interval
        return 0

    cdef Py_ssize_t free_transient(self, Attribute *attribute) noexcept nogil:
        if attribute.attribute != NULL:
            free(attribute.attribute)
        return 0

    cdef Py_ssize_t free_persistent(self, Attribute *attribute) noexcept nogil:
        if attribute.attribute != NULL:
            free(attribute.attribute)
        return 0

    cdef Py_ssize_t init_persistent(
            self,
            TSArray X,
            Attribute *transient,
            Attribute *persistent
    ) noexcept nogil:
        cdef Interval *from_interval = <Interval*> transient.attribute
        cdef Interval *to_interval = <Interval*> malloc(sizeof(Interval))
        to_interval.start = from_interval.start
        to_interval.length = from_interval.length
        to_interval.random_output = from_interval.random_output
        persistent.dim = transient.dim
        persistent.attribute = to_interval
        return 0

    cdef double transient_value(
            self,
            Attribute *attribute,
            TSArray X,
            Py_ssize_t sample
    ) noexcept nogil:
        cdef Interval *interval = <Interval*> attribute.attribute
        return self.summarizer.summarize(
            interval.random_output,
            &X[sample, attribute.dim, interval.start],
            interval.length,
        )

    cdef double persistent_value(
            self,
            Attribute *attribute,
            TSArray X,
            Py_ssize_t sample
    ) noexcept nogil:
        return self.transient_value(attribute, X, sample)

    cdef Py_ssize_t transient_fill(
            self,
            Attribute *attribute,
            TSArray X,
            Py_ssize_t sample,
            double[:, :] out,
            Py_ssize_t out_sample,
            Py_ssize_t out_attribute,
    ) noexcept nogil:
        cdef Py_ssize_t n_summarizers = self.summarizer.n_outputs()
        cdef Interval *interval = <Interval*> attribute.attribute
        self.summarizer.summarize_all(
            &X[sample, attribute.dim, interval.start],
            interval.length,
            &out[out_sample, out_attribute * n_summarizers],  # TODO: non-contig
        )
        return 0

    cdef Py_ssize_t persistent_fill(
            self,
            Attribute *attribute,
            TSArray X,
            Py_ssize_t sample,
            double[:, :] out,
            Py_ssize_t out_sample,
            Py_ssize_t out_attribute,
    ) noexcept nogil:
        return self.transient_fill(
            attribute, X, sample, out, out_sample, out_attribute
        )

    cdef object persistent_to_object(self, Attribute *attribute):
        cdef Interval *interval = <Interval*> attribute.attribute
        return attribute.dim, (interval.start, interval.length, interval.random_output)

    cdef Py_ssize_t persistent_from_object(self, object object, Attribute *attribute):
        dim, (start, length, random_output) = object
        cdef Interval *interval = <Interval*> malloc(sizeof(Interval))
        interval.start = start
        interval.length = length
        interval.random_output = random_output

        attribute.dim = dim
        attribute.attribute = interval
        return 0


cdef class RandomFixedIntervalAttributeGenerator(IntervalAttributeGenerator):

    cdef Py_ssize_t n_random_interval
    cdef Py_ssize_t *random_attribute_id

    def __cinit__(
            self,
            Py_ssize_t n_intervals,
            Summarizer summarizer,
            Py_ssize_t n_random_interval
    ):
        self.n_random_interval = n_random_interval
        self.random_attribute_id = <Py_ssize_t*> malloc(
            sizeof(Py_ssize_t) * n_intervals
        )
        cdef Py_ssize_t i
        for i in range(n_intervals):
            self.random_attribute_id[i] = i

    cdef Py_ssize_t get_n_attributess(self, TSArray X) noexcept nogil:
        return X.shape[1] * self.n_random_interval

    cdef Py_ssize_t next_attribute(
            self,
            Py_ssize_t attribute_id,
            TSArray X,
            Py_ssize_t *samples,
            Py_ssize_t n_samples,
            Attribute *transient,
            uint32_t *seed,
    ) noexcept nogil:
        # reshuffle the attribute_ids for each dimension
        if attribute_id % self.n_random_interval == 0:
            shuffle(self.random_attribute_id, self.n_intervals, seed)

        # we need to rescale the attribute_id for the correct dimension
        cdef Py_ssize_t padding = attribute_id // self.n_random_interval + 1
        return IntervalAttributeGenerator.next_attribute(
            self,
            self.random_attribute_id[attribute_id] * padding,
            X,
            samples,
            n_samples,
            transient,
            seed
        )

    def __reduce__(self):
        return self.__class__, (
            self.n_intervals,
            self.summarizer,
            self.n_random_interval
        )


cdef class RandomIntervalAttributeGenerator(IntervalAttributeGenerator):

    cdef Py_ssize_t min_length
    cdef Py_ssize_t max_length

    def __cinit__(
            self,
            Py_ssize_t n_intervals,
            Summarizer summarizer,
            Py_ssize_t min_length,
            Py_ssize_t max_length,
    ):
        self.min_length = min_length
        self.max_length = max_length

    def __reduce__(self):
        return self.__class__, (
            self.n_intervals,
            self.summarizer,
            self.min_length,
            self.max_length,
        )

    cdef Py_ssize_t get_n_attributess(self, TSArray X) noexcept nogil:
        return self.n_intervals

    cdef Py_ssize_t next_attribute(
            self,
            Py_ssize_t attribute_id,
            TSArray X,
            Py_ssize_t *samples,
            Py_ssize_t n_samples,
            Attribute *transient,
            uint32_t *seed,
    ) noexcept nogil:
        cdef Interval *interval = <Interval*> malloc(sizeof(Interval))

        interval.length = rand_int(self.min_length, self.max_length, seed)
        interval.start = rand_int(0, X.shape[2] - interval.length, seed)
        interval.random_output = 0
        if self.summarizer.n_outputs() > 1:
            interval.random_output = rand_int(0, self.summarizer.n_outputs(), seed)

        transient.dim = 0
        if X.shape[1] > 1:
            transient.dim = rand_int(1, X.shape[1], seed)

        transient.attribute = interval
        return 0



# Authors: Isak Samsten
# License: BSD 3 clause

import math
import numbers
from copy import deepcopy

import numpy as np
from sklearn.utils import check_random_state, resample
from sklearn.utils._param_validation import Interval, StrOptions
from sklearn.utils.validation import check_is_fitted

from ...base import BaseEstimator, CounterfactualMixin, ExplainerMixin
from ...distance._distance import (
    _METRICS,
    paired_distance,
    pairwise_subsequence_distance,
)
from ...ensemble._ensemble import BaseShapeletForestClassifier


def _min_euclidean_distance(shapelet, x):
    return pairwise_subsequence_distance(
        shapelet.reshape(-1),
        x.reshape(-1),
        dim=0,
        metric="euclidean",
        return_index=True,
    )


def _shapelet_transform(shapelet, x, start_index, theta):
    x_shapelet = x[start_index : (start_index + shapelet.shape[0])]
    shapelet_diff = x_shapelet - shapelet
    dist = np.linalg.norm(shapelet_diff)
    if dist == 0:
        # If the distance between the shapelets is zero, we move it in a random
        # direction towards the distance threshold.
        # TODO: ensure that this uses a consistent random_seed for repeatable runs.
        x_shapelet = np.random.uniform(shapelet.shape)
        shapelet_diff = x_shapelet - shapelet
        dist = np.linalg.norm(shapelet_diff)

    return shapelet + shapelet_diff / dist * theta


def _path_transform(x, path, epsilon):
    for direction, (dim, shapelet), threshold in path:
        if x.ndim == 2:
            x_dim = x[dim]
        else:
            x_dim = x

        dist, location = _min_euclidean_distance(shapelet, x_dim)
        if direction < 0:
            if dist > threshold:
                impute_shape = _shapelet_transform(
                    shapelet, x_dim, location, threshold - epsilon
                )
                x_dim[location : location + len(shapelet)] = impute_shape
        else:
            while dist - threshold < 0:
                impute_shape = _shapelet_transform(
                    shapelet, x_dim, location, threshold + epsilon
                )
                x_dim[location : location + len(shapelet)] = impute_shape
                dist, location = _min_euclidean_distance(shapelet, x_dim)
    return x


class PredictionPaths:
    def __init__(self, classes):
        self.classes = classes
        self._paths = {c: [] for c in classes}

    def _append(self, tree):
        left = tree.left
        right = tree.right
        threshold = tree.threshold
        shapelet = tree.attribute
        value = tree.value

        def recurse(node_id, path):
            if left[node_id] != -1:
                left_path = path.copy()
                (dim, (dim_, data)) = shapelet[node_id]
                left_path.append((-1, (dim, data), threshold[node_id]))
                recurse(left[node_id], left_path)

                right_path = path.copy()
                right_path.append((1, (dim, data), threshold[node_id]))
                recurse(right[node_id], right_path)
            else:
                self._paths[self.classes[np.argmax(value[node_id])]].append(path)

        recurse(0, [])

    def prune(self, max_paths, random_state):
        self._paths = {
            label: resample(
                paths,
                replace=False,
                n_samples=math.ceil(len(paths) * max_paths),
                random_state=random_state.randint(np.iinfo(np.int32).max),
            )
            for label, paths in self._paths.items()
        }

    def __contains__(self, item):
        return item in self._paths

    def __getitem__(self, item):
        return self._paths[item]


def _cost_wrapper(metric):
    def f(x, y, aggregation=np.mean):
        if x.ndim == 1 and y.ndim == 1:
            return paired_distance(x.reshape(1, -1), y.reshape(1, -1), metric=metric)
        elif x.ndim == 2 and y.ndim == 2:
            return paired_distance(x, y, metric=metric)
        elif x.ndim == 3 and y.ndim == 3:
            return aggregation(
                [
                    paired_distance(x, y, metric=metric, dim=i)
                    for i in range(x.shape[1])
                ],
                axis=0,
            )
        else:
            raise ValueError("invalid dim")

    return f


_AGGREGATION = {
    "median": np.median,
    "mean": np.mean,
}


class ShapeletForestCounterfactual(CounterfactualMixin, ExplainerMixin, BaseEstimator):
    """
    Counterfactual explanations for shapelet forest classifiers.

    Parameters
    ----------
    cost : {"euclidean", "cosine", "manhattan"} or callable, optional
        The cost function to determine the goodness of counterfactual.
    aggregation : callable, optional
        The aggregation function for the cost of multivariate counterfactuals.
    epsilon : float, optional
        Control the degree of change from the decision threshold.
    batch_size : float, optional
        Batch size when evaluating the cost and predictions of
        counterfactual candidates. The default setting is to evaluate
        all counterfactual samples.

        .. versionchanged:: 1.1
            The default value changed to 0.1
    max_paths : float, optional
        Sample a fraction of the positive prediction paths.

        .. versionadded:: 1.1
            Add support for subsampling prediction paths.
    verbose : bool, optional
        Print information to stdout during execution.
    random_state : RandomState or int, optional
        Pseudo-random number for consistency between different runs.

    Attributes
    ----------
    paths_ : dict
        A dictionary of prediction paths per label

    Warnings
    --------
    Only shapelet forests fit with the Euclidean distance is supported i.e.,
    ``metric="euclidean"``

    Notes
    -----
    This implementation only supports the reversible algorithm
    described by Karlsson (2020)

    References
    ----------
    Karlsson, I., Rebane, J., Papapetrou, P., & Gionis, A. (2020).
        Locally and globally explainable time series tweaking.
        Knowledge and Information Systems, 62(5), 1671-1700.

    Karlsson, I., Rebane, J., Papapetrou, P., & Gionis, A. (2018).
        Explainable time series tweaking via irreversible and reversible temporal
        transformations. In 2018 IEEE International Conference on Data Mining (ICDM)
    """

    _parameter_constraints: dict = {
        "cost": [StrOptions(_METRICS.keys()), callable],
        "aggregation": [StrOptions(_AGGREGATION.keys()), callable],
        "epsilon": [Interval(numbers.Real, 0, None, closed="right")],
        "batch_size": [Interval(numbers.Real, 0, 1, closed="right")],
        "max_paths": [Interval(numbers.Real, 0, 1, closed="right")],
        "verbose": ["verbose"],
        "random_state": ["random_state"],
    }

    def __init__(
        self,
        *,
        cost="euclidean",
        aggregation="mean",
        epsilon=1.0,
        batch_size=0.1,
        max_paths=1.0,
        verbose=False,
        random_state=None,
    ):
        self.cost = cost
        self.aggregation = aggregation
        self.epsilon = epsilon
        self.random_state = random_state
        self.batch_size = batch_size
        self.max_paths = max_paths
        self.verbose = verbose

    def _validate_estimator(self, estimator, allow_3d=False):
        if not isinstance(estimator, BaseShapeletForestClassifier):
            raise ValueError(
                "estimator must be ShapeletForestClassifier, not %r"
                % type(estimator).__qualname__
            )

        return super()._validate_estimator(estimator, allow_3d)

    def fit(self, estimator, x=None, y=None):
        self._validate_params()
        estimator = self._validate_estimator(estimator, allow_3d=True)

        self.cost_ = _cost_wrapper(self.cost)
        if isinstance(self.aggregation, str):
            self.aggregation_ = _AGGREGATION[self.aggregation]
        else:
            self.aggregation_ = self.aggregation

        self.estimator_ = deepcopy(estimator)
        self.paths_ = PredictionPaths(estimator.classes_)

        for base_estimator in self.estimator_.estimators_:
            self.paths_._append(base_estimator.tree_)

        self.epsilon_ = self.epsilon
        self.batch_size_ = self.batch_size
        self.paths_.prune(self.max_paths, check_random_state(self.random_state))
        return self

    def explain(self, x, y):
        check_is_fitted(self)
        x, y = self._validate_data(x, y, allow_3d=True, reset=False, dtype=float)
        counterfactuals = np.empty(x.shape)

        for i in range(x.shape[0]):
            if self.verbose:
                print(
                    f"Generating counterfactual for the {i}:th sample. "
                    f"The desired target label is {y[i]}."
                )

            if y[i] not in self.paths_:
                raise ValueError("unknown label, got %r" % y)

            t = self._candidates(x[i], y[i])
            if t is not None:
                counterfactuals[i] = t
            else:
                counterfactuals[i] = x[i]

        return counterfactuals

    def _candidates(self, x, y):
        prediction_paths = self.paths_[y]
        n_counterfactuals = len(prediction_paths)
        counterfactuals = np.empty((n_counterfactuals,) + x.shape)
        for i, path in enumerate(prediction_paths):
            counterfactuals[i] = _path_transform(x.copy(), path, self.epsilon_)

        # Note that the cost is ordered in increasing order; hence, if a
        # conversion is successful there can exist no other successful
        # transformation with lower cost.
        x = np.broadcast_to(x, counterfactuals.shape)
        cost_sort = np.argsort(
            self.cost_(counterfactuals, x, aggregation=self.aggregation_)
        )

        batch_size = math.ceil(n_counterfactuals * self.batch_size_)
        for i in range(0, n_counterfactuals, batch_size):
            batch_cost = cost_sort[i : min(n_counterfactuals, i + batch_size)]
            batch_counterfactuals = counterfactuals[batch_cost]
            batch_prediction = self.estimator_.predict(batch_counterfactuals)
            batch_counterfactuals = batch_counterfactuals[batch_prediction == y]
            if batch_counterfactuals.shape[0] > 0:
                return batch_counterfactuals[0]

        return None


# class IncrementalTreeLabelTransform(CounterfactualTransformer):
#     def _transform_single_path(self, x, path):
#         for direction, shapelet, threshold in path:
#             if direction == "<=":
#                 dist, location = distance(
#                     shapelet, x, metric="euclidean", return_index=True)
#                 if dist > threshold:
#                     impute_shape = _shapelet_transform(shapelet, x, location,
#                                                        threshold - self.epsilon)
#                     x[location:(location + len(shapelet))] = impute_shape
#             else:
#                 dist, location = distance(
#                     shapelet, x, metric="euclidean", return_index=True)
#                 while dist - threshold < 0.0001:
#                     impute_shape = _shapelet_transform(shapelet, x, location,
#                                                        threshold + self.epsilon)
#                     x[location:(location + len(shapelet))] = impute_shape
#                     dist, location = distance(
#                         shapelet, x, metric="euclidean", return_index=True)
#         return x
#
#
# def locked_iter(locked, start, end):
#     """iterates over unlocked regions. `locked must be sorterd`
#     :param locked: a list of tuples `[(start_lock, end_lock)]`
#     :param start: first index
#     :param end: last index
#     :yield: unlocked regions
#     """
#     if len(locked) == 0:
#         yield start, end
#     for i in range(0, len(locked)):
#         s, e = locked[i]
#         if i == 0:
#             start = 0
#         if s - start > 0:
#             yield start, (start + (s - start))
#         start = e
#
#         if i == len(locked) - 1 and end - start > 0:
#             yield start, end
#
#
# def in_range(i, start, end):
#     return start < i <= end
#
#
# def is_locked(start, end, locked):
#     """Return true if location is locked
#     :param pos: the position
#     :param locked: list of locked positions
#     :returns: true if locked
#     """
#     for s, e in locked:
#         if in_range(s, start, end) or in_range(e, start, end):
#             return True
#     return False
#
#
# class LockingIncrementalTreeLabelTransform(IncrementalTreeLabelTransform):
#     def _transform_single(self, x_orig):
#         path_list = self.paths_[self.to_label_]
#         best_cost = np.inf
#         best_x_prime = np.nan
#         batch_size = round(len(path_list) * self.batch_size) + 1
#         x_prime = np.empty((batch_size, x_orig.shape[0]))
#         n_paths = len(path_list)
#         i = 0
#         prune = 0
#         predictions = 0
#         while i < n_paths:
#             j = 0
#             while j < batch_size and i + j < n_paths:
#                 path = path_list[i + j]
#                 x_prime_j = self._transform_single_path(
#                     x_orig, path, best_cost)
#                 if x_prime_j is not None:
#                     x_prime[j, :] = x_prime_j
#                     j += 1
#                 else:
#                     prune += 1
#                     i += 1
#             i += j
#             if j > 0:
#                 predictions += j
#                 x_prime_pred = x_prime[:j, :]
#                 cond = self.ensemble_.predict(x_prime_pred) == self.to_label_
#                 x_prime_i = x_prime_pred[cond]
#                 if x_prime_i.shape[0] > 0:
#                     cost = self._compute_cost(x_prime_i, x_orig)
#                     argmin_cost = np.argmin(cost)
#                     min_cost = cost[argmin_cost]
#                     if min_cost < best_cost:
#                         best_cost = min_cost
#                         best_x_prime = x_prime_i[argmin_cost]
#
#         return (best_x_prime, predictions / float(len(path_list)),
#                 prune / float(len(path_list)))
#
#     def _transform_single_path(self, x_orig, path, best_cost=np.inf):
#         x_prime = x_orig.copy()
#         locked = []
#         for direction, shapelet, threshold in path:
#             if direction == "<=":
#                 dist, location = distance(
#                     shapelet, x_prime, metric="euclidean", return_index=True)
#
#                 if dist > threshold and not is_locked(
#                         location, location + len(shapelet), locked):
#                     impute_shape = _shapelet_transform(
#                         shapelet, x_prime, location, threshold - self.epsilon)
#
#                     x_prime[location:(location + len(shapelet))] = impute_shape
#                     locked.append((location, location + len(shapelet)))
#
#                     cost = self._compute_cost(x_prime, x_orig)
#                     if cost >= best_cost:
#                         return None
#
#             else:
#                 dist, location = distance(
#                     shapelet, x_prime, metric="euclidean", return_index=True)
#                 while (dist - threshold < 0.0001 and not is_locked(
#                         location, location + len(shapelet), locked)):
#                     impute_shape = _shapelet_transform(
#                         shapelet, x_prime, location, threshold + self.epsilon)
#
#                     x_prime[location:(location + len(shapelet))] = impute_shape
#                     locked.append((location, location + len(shapelet)))
#
#                     cost = self._compute_cost(x_prime, x_orig)
#                     if cost >= best_cost:
#                         return None
#
#                     dist, location = distance(
#                         shapelet,
#                         x_prime,
#                         metric="euclidean",
#                         return_index=True)
#         return x_prime
